#! /usr/bin/env python
#
# Connect to the NETCONF server passed on the command line while passing
# custom capabilities on the default device handler and display the
# client capabilities. For brevity and clarity of the examples, we omit
# proper exception handling.
#
# $ ./nc09.py host username password

import logging
import os
import sys
import warnings

warnings.simplefilter("ignore", DeprecationWarning)
from ncclient import manager

def demo(host, user, password):
    with manager.connect(
            host=host, port=830, username=user, password=password,
            nc_params={
                'capabilities': [
                    'urn:custom:capability:1.0',
                    'urn:custom:capability:2.0'
                ]},
            hostkey_verify=False) as m:

        for c in m.client_capabilities:
            print(c)

if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)
    demo(sys.argv[1], sys.argv[2], sys.argv[3])



#! /usr/bin/env python
#
# Delete an existing user from the running configuration using
# edit-config and the test-option provided by the :validate
# capability.
#
# $ ./nc05.py broccoli bob

import sys, os, warnings
warnings.simplefilter("ignore", DeprecationWarning)
from ncclient import manager

def demo(host, user, name):
    snippet = """<nc:config xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
      <aaa xmlns="http://tail-f.com/ns/aaa/1.1">
        <authentication> <users> <user nc:operation="delete">
        <name>%s</name>
      </user></users></authentication></aaa></nc:config>""" % name

    with manager.connect(host=host, port=22, username=user) as m:
        assert(":validate" in m.server_capabilities)
        m.edit_config(target='running', config=snippet,
                      test_option='test-then-set')

if __name__ == '__main__':
    demo(sys.argv[1], os.getenv("USER"), sys.argv[2])



#!/usr/bin/env python
import logging

from ncclient import manager
from ncclient.xml_ import *


def connect(host, port, user, password):
    with manager.connect(
        host=host,
        port=port,
        username=user,
        password=password,
        timeout=60,
        device_params={'name': 'junos'},
        hostkey_verify=False
    ) as conn:
        vpns = conn.command(
            'show l2circuit connections'
        ).xpath(
            'l2circuit-connection-information/l2circuit-neighbor'
        )
        connection_dict = {}
        for vpn in vpns:
            neighbor = ''
            for tag in vpn.getchildren():
                connection_details_dict = {}
                if tag.tag == 'neighbor-address':
                    neighbor = tag.text
                if tag.tag == 'connection':
                    for child in tag.getchildren():
                        if child.tag == 'local-interface':
                            ifce = child.find('interface-name').text
                            connection_details_dict.update({'interface': ifce})
                        else:
                            if child.tag == 'connection-id':
                                cid = child.text
                            else:
                                connection_details_dict.update({child.tag: child.text})
                    connection_details_dict.update({'neighbor': neighbor})
                    if connection_dict.get(cid):
                        connection_dict.get(cid).append(connection_details_dict)
                    else:
                        connection_dict.update({cid: [connection_details_dict]})

    return connection_dict


if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    response = connect('router', 22, 'netconf', 'juniper!')
    logging.info("Response: %s", response)



# Copyright (c) Siemens AG, 2022
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket
import sys
import threading

from io import BytesIO as StringIO
from socket import AF_INET, SOCK_STREAM
from ssl import CERT_REQUIRED, SSLContext, SSLError

from ncclient.capabilities import Capabilities
from ncclient.logging_ import SessionLoggerAdapter
from ncclient.transport.errors import TLSError
from ncclient.transport.session import Session
from ncclient.transport.parser import DefaultXMLParser

logger = logging.getLogger("ncclient.transport.tls")

DEFAULT_TLS_NETCONF_PORT = 6513
DEFAULT_TLS_TIMEOUT = 120

BUF_SIZE = 4096


class TLSSession(Session):

    def __init__(self, device_handler):
        capabilities = Capabilities(device_handler.get_capabilities())
        Session.__init__(self, capabilities)
        self._host = None
        self._connected = False
        self._socket = None
        self._buffer = StringIO()
        self._device_handler = device_handler
        self._message_list = []
        self._closing = threading.Event()
        self.parser = DefaultXMLParser(self)
        self.logger = SessionLoggerAdapter(logger, {'session': self})

    def _dispatch_message(self, raw):
        self.logger.info("Received message from host")
        self.logger.debug("Received:\n%s", raw)
        return super(TLSSession, self)._dispatch_message(raw)

    def close(self):
        self._closing.set()
        self._socket.close()
        self._connected = False

    def connect(self, host=None, port=DEFAULT_TLS_NETCONF_PORT,
                keyfile=None, certfile=None, ca_certs=None,
                protocol=None, check_hostname=True, server_hostname=None,
                timeout=DEFAULT_TLS_TIMEOUT):
        """Establish NETCONF session via TLS.

        :param host: Hostname or IP address to connect to.
        :param port: Port number which would be used for connection, the
            default value is 6513 (`DEFAULT_TLS_NETCONF_PORT`).
        :param certfile: Points to file in PEM format with client's certificate
            and optionally with client's private key.
        :param keyfile: Points to file with client's private key in PEM format,
            if the latter is not in certfile.
        :param ca_certs: Points to file containing a list of certificates,
            which should be used to verify server certificate.
        :param protocol: Protocol version to use. Should be either
            ssl.PROTOCOL_TLS_CLIENT for Python 3.6+, or ssl.PROTOCOL_TLS
            for older versions.
        :param check_hostname: If set to True, perform verification of the
            hostname during the handshake.
        :param server_hostname: If set, is used upon hostname checking instead
            of the `host` parameter.
        :param timeout: Specifies the connection timeout, defaults to
            120 seconds (`DEFAULT_TLS_TIMEOUT`).

        :raise TLSError if the connection can not be established.
        """
        if host is None:
            raise TLSError('Missing host')
        if certfile is None:
            raise TLSError('Missing client certificate file')
        if protocol is None:
            raise TLSError('Missing TLS protocol')

        ssl_context = SSLContext(protocol)
        ssl_context.verify_mode = CERT_REQUIRED
        ssl_context.check_hostname = check_hostname
        try:
            ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)
        except SSLError:
            raise TLSError('Bad client private key / certificate pair')
        except IOError:
            raise TLSError('Private key / certificate pair not found')

        if ca_certs:
            try:
                ssl_context.load_verify_locations(cafile=ca_certs)
            except SSLError:
                raise TLSError('Bad Certification Authority file')
            except IOError:
                raise TLSError('CA certificate file not found')

        sock = socket.socket(AF_INET, SOCK_STREAM)
        ssl_sock = ssl_context.wrap_socket(
            sock, do_handshake_on_connect=False,
            server_hostname=server_hostname or host)
        ssl_sock.settimeout(timeout)

        try:
            ssl_sock.connect((host, port))
        except Exception:
            raise TLSError("Could not connect to %s:%s" % (host, port))

        try:
            ssl_sock.do_handshake()
        except Exception:
            raise TLSError("Unsuccessful TLS handshake with %s:%s" % (host, port))

        self._host = host
        self._socket = ssl_sock
        self._connected = True
        self._post_connect()

    def _transport_read(self):
        return self._socket.recv(BUF_SIZE)

    def _transport_write(self, data):
        return self._socket.send(data)

    def _transport_register(self, selector, event):
        selector.register(self._socket, event)

    def _send_ready(self):
        # In contrast to Paramiko's `Channel`, pure python sockets do not
        # expose `send_ready()` function.
        return True

    @property
    def host(self):
        """Host this session is connected to, or None if not connected."""
        if hasattr(self, '_host'):
            return self._host
        return None



from lxml import etree
from ncclient.xml_ import *
from ncclient.operations.rpc import RPC


class DisplayCommand(RPC):
    def request(self, cmds):
        """
        Single Execution element is permitted.
        cmds can be a list or single command
        """
        if isinstance(cmds, list):
            cmd = '\n'.join(cmds)
        elif isinstance(cmds, str) or isinstance(cmds, unicode):
            cmd = cmds

        node = etree.Element(qualify('CLI', BASE_NS_1_0))

        etree.SubElement(node, qualify('Execution',
                                       BASE_NS_1_0)).text = cmd

        return self._request(node)


class ConfigCommand(RPC):
    def request(self, cmds):
        """
        Single Configuration element is permitted.
        cmds can be a list or single command
        commands are pushed to the switch in this method
        """

        if isinstance(cmds, list):
            cmd = '\n'.join(cmds)
        elif isinstance(cmds, str) or isinstance(cmds, unicode):
            cmd = cmds

        node = etree.Element(qualify('CLI', BASE_NS_1_0))

        etree.SubElement(node, qualify('Configuration',
                                       BASE_NS_1_0)).text = cmd

        return self._request(node)


class Action(RPC):
    def request(self, action=None):
        node = new_ele("action")
        node.append(validated_element(action))
        return self._request(node)


class Save(RPC):
    def request(self, filename=None):
        node = new_ele('save')
        sub_ele(node, 'file').text = filename
        return self._request(node)


class Rollback(RPC):
    def request(self, filename=None):
        node = new_ele('rollback')
        sub_ele(node, 'file').text = filename
        return self._request(node)



#! /usr/bin/env python
#
# Retrieve the running config from the NETCONF server passed on the
# command line using get-config and write the XML configs to files.
#
# $ ./nc02.py broccoli

import sys, os, warnings
warnings.simplefilter("ignore", DeprecationWarning)
from ncclient import manager

def demo(host, user):
    with manager.connect(host=host, port=22, username=user, hostkey_verify=False) as m:
        c = m.get_config(source='running').data_xml
        with open("%s.xml" % host, 'w') as f:
            f.write(c)

if __name__ == '__main__':
    demo(sys.argv[1], os.getenv("USER"))



from ncclient.operations.session import *
import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from xml.etree import ElementTree


class TestSession(unittest.TestCase):

    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'junos'})

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_close_session(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = CloseSession(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele("close-session")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.RPC._request')
    def test_kill_session(self, mock_request, mock_session):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = KillSession(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request("100")
        node = new_ele("kill-session")
        sub_ele(node, "session-id").text = "100"
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)



import base64
import hashlib
import logging
import os
import socket
import threading
from io import BytesIO as StringIO
from typing import Any, Callable

from ncclient.capabilities import Capabilities
from ncclient.logging_ import SessionLoggerAdapter
from ncclient.transport.errors import AuthenticationError, SSHError, SSHUnknownHostError
from ncclient.transport.parser import DefaultXMLParser
from ncclient.transport.session import Session
from ssh.channel import Channel
from ssh.exceptions import AuthenticationDenied, ChannelOpenFailure, KeyImportError
from ssh.exceptions import SSHError as LibSSHError
from ssh.key import import_privkey_file, import_privkey_base64
from ssh.options import HOST, KNOWNHOSTS, USER
from ssh.session import Session as LSession

logger = logging.getLogger("ncclient.transport.libssh")

PORT_NETCONF_DEFAULT = 830
BUF_SIZE = 4096


def default_unknown_host_cb(host: str, fingerprint: str) -> bool:
    """
    An unknown host callback returns True if it finds the key acceptable, and False if not. This default
    callback always returns False, which would lead to connect() raising a SSHUnknownHost exception. Supply
    another valid callback if you need to verify the host key programmatically.

    host - the hostname or IP address of the host
    fingerprint - hex string representing the host key fingerprint,
        colon-delimited e.g. "4b:69:6c:72:6f:79:20:77:61:73:20:68:65:72:65:21"
    """
    return False


class LibSSHSession(Session):

    _buffer: StringIO
    _channel: Channel | None
    _closing: threading.Event
    _connected: bool
    _device_handler: Any
    _host: str | None
    _message_list: list
    _receiver_thread: threading.Thread | None
    _session: LSession | None
    _socket: socket.socket | None
    _socket_r: socket.socket
    _socket_w: socket.socket
    logger: SessionLoggerAdapter
    parser: DefaultXMLParser

    def __init__(self, device_handler):
        capabilities = Capabilities(device_handler.get_capabilities())
        Session.__init__(self, capabilities)
        self._host = None
        self._connected = False
        self._socket = None
        self._channel = None
        self._session = None
        self._buffer = StringIO()
        self._device_handler = device_handler
        self._message_list = []
        self._closing = threading.Event()
        self.parser = DefaultXMLParser(self)
        self.logger = SessionLoggerAdapter(logger, {"session": self})
        self._socket_r, self._socket_w = socket.socketpair()
        self._receiver_thread = None

    def _receiver_loop(self, socket_w: socket.socket, channel):
        """
        The receiver loop reads data from the SSH channel and writes it to the transport socket.

        ### Parameters:
        - socket_w: The writable end of the socket pair used for communication.
        - channel: The SSH channel to read data from.
        """
        while True:
            try:
                data = channel.read()
            except LibSSHError:
                logger.error("Channel read error")
                break
            if data[0] == 0:
                # EOF received
                self._closing.set()
                break
            socket_w.send(data[1])
        socket_w.close()

    def connect(
        self,
        host: str,
        port: int = PORT_NETCONF_DEFAULT,
        username: str | None = None,
        password: str | None = None,
        key_filename: str | None = None,
        key_base64: bytes | None = None,
        key_passphrase: bytes | None = None,
        allow_agent: bool = True,
        hostkey_verify: bool = True,
        hostkey_b64: str | None = None,
        timeout: int | None = None,
        unknown_host_cb: Callable = default_unknown_host_cb,
        bind_addr: str | None = None,
        sock: socket.socket | None = None,
        sock_fd: int | None = None,
        **kwargs,
    ):
        """
        Connect to the SSH server and establish a NETCONF session.

        Note: Some arguments are not implemented due to limitations in the underlying libssh library
        in comparison to original paramiko library. These arguments are ignored.

        ### Parameters:
        - host: The hostname or IP address of the SSH server.
        - port: The port number to connect to (default is 830).
        - username: The username for authentication.
        - password: The password for authentication.
        - key_filename: Path to the private key file for public key authentication.
        - key_base64: Base64-encoded private key for public key authentication.
        - key_passphrase: Passphrase for the private key, if required.
        - allow_agent: Whether to use the SSH agent for authentication.
        - hostkey_verify: Whether to verify the server's host key.
        - hostkey_b64: Base64-encoded public key of the server for host key verification.
        - timeout: Timeout for the connection attempt.
        - unknown_host_cb: Callback function for handling unknown hosts.
        - bind_addr: Local address to bind the socket to.
        - sock: An existing socket object to use for the connection.
        - sock_fd: An existing socket file descriptor to use for the connection.
        """
        # not supported warnings
        for arg in kwargs.keys():
            self.logger.warning(f"Argument '{arg} is not supported.")

        if not (host or sock_fd or sock):
            raise SSHError("Missing host, socket or socket fd")

        if sock_fd is None and sock is None:
            for i in socket.getaddrinfo(
                host, port, socket.AF_UNSPEC, socket.SOCK_STREAM
            ):
                af, socktype, proto, _, sa = i
                try:
                    sock = socket.socket(af, socktype, proto)
                except socket.error:
                    continue
                try:
                    if bind_addr:
                        sock.bind((bind_addr, 0))
                    sock.settimeout(timeout)
                    sock.connect(sa)
                except socket.error:
                    sock.close()
                    continue
                break

            else:
                raise SSHError(f"Could not open socket to {host}:{port}")

        elif sock_fd is not None:
            print(sock_fd)
            sock = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)

        if not username:
            username = os.getenv("USER") or os.getenv("LOGNAME") or "root"
        if not username:
            raise AuthenticationError(
                "Username must be provided or set in environment variables"
            )

        self._session = LSession()
        assert self._session is not None

        self._session.options_set(HOST, host)
        self._session.options_set(USER, username)
        self._session.options_set(KNOWNHOSTS, os.path.expanduser("~/.ssh/known_hosts"))
        self._session.options_set_port(port)
        self._session.set_socket(sock)
        self._session.connect()

        if hostkey_verify:
            is_known_host = False
            server_pubkey = self._session.get_server_publickey()
            server_pubkey_base64 = server_pubkey.export_pubkey_base64().decode()
            if hostkey_b64 is not None:
                # Check if the provided hostkey matches the server's public key
                if server_pubkey_base64 == hostkey_b64:
                    is_known_host = True
            else:
                # check known hosts file
                if self._session.is_server_known():
                    is_known_host = True

            def openssh_fingerprint_md5(base64_key: str) -> str:
                """
                Calculate md5 fingerprint of the base64-encoded public key.

                ### Parameters:
                - base64_key: Base64-encoded public key string.

                ### Returns:
                - Fingerprint as a colon-separated hex string.
                """
                key_bytes = base64.b64decode(base64_key)
                digest = hashlib.md5(key_bytes).hexdigest()
                fingerprint = ":".join(
                    digest[i : i + 2] for i in range(0, len(digest), 2)
                )
                return fingerprint

            if not is_known_host:
                fingerprint = openssh_fingerprint_md5(server_pubkey_base64)
                if not unknown_host_cb(host, fingerprint):
                    # If the host key is not known, raise an error
                    raise SSHUnknownHostError(host, fingerprint)

        def auth(
            sess: LSession,
            user: str,
            passwd: str | None,
            keyfile: str | None,
            keybase64: bytes | None,
            agent: bool,
            passphrase: bytes | None,
        ):
            """
            Authenticate the session using the provided credentials.

            ### Parameters:
            - sess: The SSH session object.
            - user: Username.
            - passwd: Password.
            - keyfile: Path to the private key file.
            - agent: Whether to use the SSH agent for authentication.
            - keybase64: Base64-encoded private key.
            - passphrase: Passphrase for the private key, if required.
            """
            if passphrase is None:
                passphrase = b""

            if user and passwd:
                try:
                    sess.userauth_password(user, passwd)
                    self.logger.debug(
                        "Password authentication successful for user %s", user
                    )
                    return
                except AuthenticationDenied:
                    self.logger.debug(
                        "Password authentication failed for user %s", user
                    )

            if keyfile:
                try:
                    privkey = import_privkey_file(keyfile)
                except KeyImportError:
                    self.logger.debug(
                        "Could not import private key from file %s", keyfile
                    )
                else:
                    try:
                        sess.userauth_publickey(privkey)
                        self.logger.debug(
                            "Public key authentication successful for user %s", user
                        )
                        return
                    except AuthenticationDenied:
                        self.logger.debug(
                            "Public key authentication failed for user %s", user
                        )

            if keybase64:
                try:
                    privkey = import_privkey_base64(keybase64, passphrase)
                except KeyImportError:
                    self.logger.debug("Could not import public key from base64 string")
                else:
                    try:
                        sess.userauth_publickey(privkey)
                        self.logger.debug(
                            "Public key authentication successful for user %s", user
                        )
                        return
                    except AuthenticationDenied:
                        self.logger.debug(
                            "Public key authentication failed for user %s", user
                        )

            if agent:
                try:
                    sess.userauth_agent(user)
                    self.logger.debug(
                        "Agent authentication successful for user %s", user
                    )
                    return
                except AuthenticationDenied:
                    self.logger.debug("Agent authentication failed for user %s", user)

            raise AuthenticationError(
                f"Authentication failed for user {user} on host {host}:{port}"
            )

        auth(
            self._session,
            username,
            password,
            key_filename,
            key_base64,
            allow_agent,
            key_passphrase,
        )

        try:
            self._channel = self._session.channel_new()
        except ChannelOpenFailure as e:
            raise SSHError(f"Could not open channel to {host}:{port}") from e

        assert self._channel is not None
        try:
            self._channel.open_session()
            self._channel.request_subsystem("netconf")
        except LibSSHError as e:
            handle_exception = self._device_handler.handle_connection_exceptions(self)
            if not handle_exception:
                raise SSHError(
                    f"Could not open NETCONF session on channel to {host}:{port}"
                ) from e

        self._receiver_thread = threading.Thread(
            target=self._receiver_loop,
            args=(self._socket_w, self._channel),
            daemon=True,
        )
        self._receiver_thread.start()

        self._host = host
        self._socket = sock
        self._connected = True
        self._post_connect()

    def _send_ready(self):
        """
        Check if the session is ready to send data.

        ### Returns:
        - True if the session is connected and thus the channel is open, False otherwise.
        """
        assert self._channel is not None
        return self._channel.is_open()

    def _transport_read(self):
        """
        Read data from the transport layer.

        ### Returns:
        - Data read from the transport layer.
        """
        return self._socket_r.recv(BUF_SIZE)

    def _transport_write(self, data: bytes) -> int:
        """
        Write data to the transport layer.

        ### Parameters:
        - data: The data to be sent over the transport layer.

        ### Returns:
        - The number of bytes written to the transport layer.
        """
        assert self._channel is not None
        res = self._channel.write(data)
        return res[0]

    def _transport_register(self, selector, event):
        """
        Register the socket with the selector for the specified event.

        ### Parameters:
        - selector: The selector to register the socket with.
        - event: The event type to register for (e.g., read, write).
        """
        selector.register(self._socket_r, event)

    def close(self):
        """
        Close the SSH session and clean up resources.
        """
        assert self._channel is not None
        assert self._socket is not None
        self._closing.set()
        if not self._channel.is_closed():
            self._channel.close()
        # wait for the transport thread to close.
        if self.is_alive() and (self is not threading.current_thread()):
            self.join()
        self._channel = None
        self._socket.close()
        self._connected = False

    @property
    def host(self):
        """
        Host this session is connected to, or None if not connected.
        """
        if self._connected:
            return self._host
        return None



"""Version compatibility shim for ncclient.

This keeps the historical ``get_versions()['version']`` API while using
hatch-vcs as the build backend source of truth.
"""

from __future__ import annotations

from importlib import metadata
from pathlib import Path
import re
import subprocess
from typing import Any, Dict, Optional


def _from_installed_metadata() -> Optional[str]:
    try:
        return metadata.version("ncclient")
    except metadata.PackageNotFoundError:
        return None


def _from_git_describe() -> str:
    root = Path(__file__).resolve().parents[1]
    try:
        described = subprocess.check_output(
            [
                "git",
                "describe",
                "--tags",
                "--long",
                "--dirty",
                "--always",
                "--match",
                "v[0-9]*",
            ],
            cwd=str(root),
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
    except Exception:
        return "0+unknown"

    match = re.match(
        r"^v(?P<tag>\d+(?:\.\d+)*)((?:-(?P<distance>\d+)-g(?P<sha>[0-9a-f]+))?)(?P<dirty>-dirty)?$",
        described,
    )
    if not match:
        return "0+unknown"

    tag = match.group("tag")
    distance = match.group("distance")
    sha = match.group("sha")
    dirty = match.group("dirty")

    if distance is None or sha is None:
        return tag

    local = f"{distance}.g{sha}"
    if dirty:
        local = f"{local}.dirty"
    return f"{tag}+{local}"


def get_versions() -> Dict[str, Any]:
    version = _from_installed_metadata() or _from_git_describe()
    return {
        "version": version,
        "full-revisionid": None,
        "dirty": ".dirty" in version,
        "error": None,
        "date": None,
    }



import unittest
from ncclient.devices.ciena import *
from ncclient.xml_ import *


class TestCienaDevice(unittest.TestCase):
    
    def setUp(self):
        self.obj = CienaDeviceHandler({'name': 'ciena'})

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected["nsmap"] = self.obj.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())



import unittest
from unittest.mock import patch

import paramiko

from ncclient.devices.junos import *
import ncclient.transport

xml = '''<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
        <xsl:output method="xml" indent="no"/>

        <xsl:template match="/|comment()|processing-instruction()">
            <xsl:copy>
                <xsl:apply-templates/>
            </xsl:copy>
        </xsl:template>

        <xsl:template match="*">
            <xsl:element name="{local-name()}">
                <xsl:apply-templates select="@*|node()"/>
            </xsl:element>
        </xsl:template>

        <xsl:template match="@*">
            <xsl:attribute name="{local-name()}">
                <xsl:value-of select="."/>
            </xsl:attribute>
        </xsl:template>
        </xsl:stylesheet>
        '''
xml2 = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos">
<routing-engine>
<name>reX</name>
<commit-success/>
<ok/>
</rpc-reply>"""

xml3 = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos">
<routing-engine>
<name>reX</name>
<commit-success/>
<routing-engine/>
<ok/>
</rpc-reply>"""

xml4 = """<rpc-reply>
<operation>commit</operation>
<name>reX</name>
<commit-success/>
<ok/>
</rpc-reply>
<rpc-error>
<message>commit failure</message>
</rpc-error>
<hello>greeting!</hello>"""


class TestJunosDevice(unittest.TestCase):

    def setUp(self):
        self.obj = JunosDeviceHandler({'name': 'junos'})

    @patch('paramiko.Channel.exec_command')
    @patch('paramiko.Transport.__init__')
    @patch('paramiko.Transport.open_channel')
    def test_handle_connection_exceptions(
            self, mock_open, mock_init, mock_channel):
        session = ncclient.transport.SSHSession(self.obj)
        session._channel_id = 100
        mock_init.return_value = None
        session._transport = paramiko.Transport()
        channel = paramiko.Channel(100)
        mock_open.return_value = channel
        self.obj.handle_connection_exceptions(session)
        self.assertEqual(channel._name, "netconf-command-100")
        self.assertEqual(
            mock_channel.call_args_list[0][0][0],
            "xml-mode netconf need-trailer")

    def test_additional_operations(self):
        dict = {}
        dict["rpc"] = ExecuteRpc
        dict["get_configuration"] = GetConfiguration
        dict["load_configuration"] = LoadConfiguration
        dict["compare_configuration"] = CompareConfiguration
        dict["command"] = Command
        dict["reboot"] = Reboot
        dict["halt"] = Halt
        dict["commit"] = Commit
        dict["rollback"] = Rollback
        self.assertEqual(dict, self.obj.add_additional_operations())

    def test_transform_reply(self):
        reply = xml.encode('utf-8')
        self.assertEqual(self.obj.transform_reply(), reply)

    def test_perform_quality_check(self):
        self.assertFalse(self.obj.perform_qualify_check())

    def test_handle_raw_dispatch(self):
        self.assertFalse(self.obj.handle_raw_dispatch(xml))

        expected = re.sub(r'<ok/>', '</routing-engine>\n<ok/>', xml3)
        self.assertEqual(expected, self.obj.handle_raw_dispatch(xml3))

        expected = 'undefined: not an error message in the reply. Enable debug'
        self.assertEqual(expected, str(self.obj.handle_raw_dispatch(xml4)))



import unittest
from unittest.mock import patch

from ncclient.operations.rpc import *
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from ncclient.capabilities import Capabilities
from xml.sax.saxutils import escape

patch_str = 'ncclient.operations.rpc.Event.is_set'

xml1 = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos">
    <ok/>
</rpc-reply>"""

xml2 = """<rpc-reply message-id="urn:uuid:15ceca00-904e-11e4-94ad-5c514f91ab3f">
	<load-configuration-results>
		<rpc-error>
			<error-severity>error</error-severity>
			<error-info>
				<bad-element>system1</bad-element>
			</error-info>
			<error-message>syntax error</error-message>
		</rpc-error>
		<rpc-error>
			<error-severity>error</error-severity>
			<error-info>
				<bad-element>}</bad-element>
			</error-info>
            <error-message>error recovery ignores input until this point</error-message>
		</rpc-error>
	</load-configuration-results>
</rpc-reply>"""

xml3 = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos" attrib1 = "test">
    <software-information>
        <host-name>R1</host-name>
        <product-model>firefly-perimeter</product-model>
        <product-name>firefly-perimeter</product-name>
        <package-information>
            <name>junos</name>
            <comment>JUNOS Software Release [12.1X46-D10.2]</comment>
        </package-information>
    </software-information>
    <rpc-error>
    </rpc-error>
    <cli>
        <banner></banner>
    </cli>
</rpc-reply>"""

xml4 = """<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/14.2I0/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:b19400d6-fa2a-11e4-8f7b-0800278ff596">
<data>
<configuration xmlns="http://xml.juniper.net/xnm/1.1/xnm" junos:changed-seconds="1431599824" junos:changed-localtime="2015-05-14 03:37:04 PDT">
    <system>
        <services>
            <ftp>
                <connection-limit>200</connection-limit>
                <rate-limit>214</rate-limit>
            </ftp>
            <ssh>
                <root-login>allow</root-login>
            </ssh>
            <telnet>
            </telnet>
            <netconf>
                <ssh>
                </ssh>
            </netconf>
        </services>
    </system>
</configuration>
</data>
</rpc-reply>"""

# Huge dummy text configuration to trigger "xml.etree.XMLSyntaxError: xmlSAX2Characters: huge text node"
# - The maximum size of a single text node is `#define XML_MAX_TEXT_LENGTH 10000000`
#   (https://gitlab.gnome.org/GNOME/libxml2/blob/master/include/libxml/parserInternals.h)
# - The text should contain at least one special character that should be escaped (e.g. '>')
huge_configuration_text = "-->\n" + "Huge configuration. " * 500000

xml5_huge = """<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/18.2R1/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:b8b20a95-7822-4c59-97a3-44435d6fc6a7">
<configuration-text xmlns="http://xml.juniper.net/xnm/1.1/xnm">%s</configuration-text>
</rpc-reply>""" % escape(huge_configuration_text)

xml6 = """<rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
			<error-type>application</error-type>
			<error-tag>invalid-value</error-tag>
			<error-severity>error</error-severity>
			<error-path>path/to/node</error-path>
			<error-info>
				<bad-element>system1</bad-element>
			</error-info>
			<error-app-tag>app-tag1</error-app-tag>
			<error-message>syntax error</error-message>
	  </rpc-error>
"""

xml7 = """<rpc-error xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">
			<error-type>protocol</error-type>
			<error-tag>missing-element</error-tag>
			<error-severity>error</error-severity>
			<error-path>path/to/different/node</error-path>
			<error-info>
				<bad-element>system2</bad-element>
			</error-info>
			<error-app-tag>app-tag2</error-app-tag>
			<error-message>missing element error</error-message>
	  </rpc-error>
"""

class TestRPC(unittest.TestCase):

    def test_rpc_reply(self):
        obj = RPCReply(xml4)
        obj.parse()
        self.assertTrue(obj.ok)
        self.assertFalse(obj.error)
        self.assertEqual(xml4, obj.xml)
        self.assertTrue(obj._parsed)

    def test_rpc_reply_huge_text_node_exception(self):
        obj = RPCReply(xml5_huge)
        self.assertRaises(etree.XMLSyntaxError, obj.parse)

    def test_rpc_reply_huge_text_node_workaround(self):
        obj = RPCReply(xml5_huge, huge_tree=True)
        obj.parse()
        self.assertTrue(obj.ok)
        self.assertFalse(obj.error)
        self.assertEqual(xml5_huge, obj.xml)
        self.assertTrue(obj._parsed)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_rpc_send(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)
        reply = RPCReply(xml1)
        obj._reply = reply
        node = new_ele("commit")
        sub_ele(node, "confirmed")
        sub_ele(node, "confirm-timeout").text = "50"
        sub_ele(node, "log").text = "message"
        result = obj._request(node)
        ele = new_ele("rpc",
                      {"message-id": obj._id},
                      **device_handler.get_xml_extra_prefix_kwargs())
        ele.append(node)
        node = to_xml(ele)
        mock_send.assert_called_once_with(node)
        self.assertEqual(
            result.data_xml,
            (NCElement(
                reply,
                device_handler.transform_reply())).data_xml)
        self.assertEqual(obj.session, session)
        self.assertEqual(reply, obj.reply)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_generic_rpc_send(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = GenericRPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)
        reply = RPCReply(xml1)
        obj._reply = reply
        rpc_command = 'edit-config'
        filters = ('subtree', '<top xmlns="urn:mod1"/>')

        result = obj.request(rpc_command, source='running', target='candidate', filter=filters)
        ele = new_ele("rpc",
                      {"message-id": obj._id},
                      **device_handler.get_xml_extra_prefix_kwargs())
        child = new_ele(rpc_command)
        child.append(util.datastore_or_url('target', 'candidate'))
        child.append(util.datastore_or_url('source', 'running'))
        child.append(util.build_filter(filters))

        ele.append(child)
        node = to_xml(ele)
        mock_send.assert_called_once_with(node)
        self.assertEqual(
            result.data_xml,
            (NCElement(
                reply,
                device_handler.transform_reply())).data_xml)
        self.assertEqual(obj.session, session)
        self.assertEqual(reply, obj.reply)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_rpc_async(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(
            session,
            device_handler,
            raise_mode=RaiseMode.ALL,
            timeout=0,
            async_mode=True)
        reply = RPCReply(xml1)
        obj._reply = reply
        node = new_ele("commit")
        result = obj._request(node)
        self.assertEqual(result, obj)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_rpc_timeout_error(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)
        reply = RPCReply(xml1)
        obj.deliver_reply(reply)
        node = new_ele("commit")
        sub_ele(node, "confirmed")
        mock_thread.return_value = False
        self.assertRaises(TimeoutExpiredError, obj._request, node)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_rpc_rpcerror(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)
        reply = RPCReply(xml1)
        obj._reply = reply
        node = new_ele("commit")
        sub_ele(node, "confirmed")

        err = RPCError(to_ele(xml2))
        obj.deliver_error(err)
        self.assertRaises(RPCError, obj._request, node)

    def test_rpc_rpcerror_tag_to_attr(self):
        '''All elements in <rpc-error> extracted.'''
        err = RPCError(to_ele(xml6))

        self.assertEqual(None, err.errlist)

        self.assertEqual("application", err.type)
        self.assertEqual("invalid-value", err.tag)
        self.assertEqual("error", err.severity)
        self.assertEqual("path/to/node", err.path)
        self.assertEqual("app-tag1", err.app_tag)
        self.assertEqual("syntax error", err.message)
        self.assertIn("<bad-element>system1</bad-element>", err.info)

    def test_rpc_rpcerror_multiple_errors(self):
        '''Multiple errors in <rpc-reply> extracted correctly'''
        errlist = [RPCError(to_ele(xml6)), RPCError(to_ele(xml7))]

        multiple_xml = (
        '<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">' +
        xml6 + xml7 +
        "</rpc-reply>")

        multiple_err = RPCError(to_ele(multiple_xml), errs=errlist)
        errs = multiple_err.errlist

        self.assertEqual(2, len(errs))

        self.assertEqual("application", errs[0].type)
        self.assertEqual("invalid-value", errs[0].tag)
        self.assertEqual("error", errs[0].severity)
        self.assertEqual("path/to/node", errs[0].path)
        self.assertEqual("app-tag1", errs[0].app_tag)
        self.assertEqual("syntax error", errs[0].message)
        self.assertIn("<bad-element>system1</bad-element>", errs[0].info)

        self.assertEqual("protocol", errs[1].type)
        self.assertEqual("missing-element", errs[1].tag)
        self.assertEqual("error", errs[1].severity)
        self.assertEqual("path/to/different/node", errs[1].path)
        self.assertEqual("app-tag2", errs[1].app_tag)
        self.assertEqual("missing element error", errs[1].message)
        self.assertIn("<bad-element>system2</bad-element>", errs[1].info)

    @patch('ncclient.transport.Session.send')
    @patch(patch_str)
    def test_rpc_capability_error(self, mock_thread, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        session._server_capabilities = [':running']
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)
        obj._assert(':running')
        self.assertRaises(MissingCapabilityError,
            obj._assert, ':candidate')

    @patch('ncclient.transport.Session.send')
    def test_rpc_huge_text_node_exception(self, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0)

        obj.deliver_reply(xml5_huge)
        node = new_ele("get-configuration", {'format': 'text'})
        self.assertRaises(etree.XMLSyntaxError, obj._request, node)

    @patch('ncclient.transport.Session.send')
    def test_rpc_huge_text_node_workaround(self, mock_send):
        device_handler, session = self._mock_device_handler_and_session()
        obj = RPC(session, device_handler, raise_mode=RaiseMode.ALL, timeout=0, huge_tree=True)
        self.assertTrue(obj.huge_tree)

        obj.deliver_reply(xml5_huge)
        node = new_ele("get-configuration", {'format': 'text'})
        result = obj._request(node)

        self.assertEqual(result.find('configuration-text').text, huge_configuration_text)
        obj.huge_tree = False
        self.assertFalse(obj.huge_tree)

    def _mock_device_handler_and_session(self):
        device_handler = manager.make_device_handler({'name': 'junos'})
        capabilities = Capabilities(device_handler.get_capabilities())
        session = ncclient.transport.Session(capabilities)
        return device_handler, session



# Copyright 2009 Shikhar Bhushan
# Copyright 2011 Leonidas Poulopoulos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Methods for creating, parsing, and dealing with XML and ElementTree objects."


import io
import types
from io import BytesIO, StringIO

from lxml import etree

# In case issues come up with XML generation/parsing
# make sure you have the ElementTree v1.2.7+ lib as
# well as lxml v3.0+

from ncclient import NCClientError

parser = etree.XMLParser(recover=False)
huge_parser = etree.XMLParser(recover=False, huge_tree=True)


def _get_parser(huge_tree=False):
    return huge_parser if huge_tree else parser


class XMLError(NCClientError):
    pass

### Namespace-related

#: Base NETCONF namespace
BASE_NS_1_0 = "urn:ietf:params:xml:ns:netconf:base:1.0"
#: YANG (RFC 6020/RFC 7950) namespace
YANG_NS_1_0 = "urn:ietf:params:xml:ns:yang:1"
#: NXOS_1_0
NXOS_1_0 = "http://www.cisco.com/nxos:1.0"
#: NXOS_IF
NXOS_IF = "http://www.cisco.com/nxos:1.0:if_manager"
#: Namespace for Tail-f core data model
TAILF_AAA_1_1 = "http://tail-f.com/ns/aaa/1.1"
#: Namespace for Tail-f execd data model
TAILF_EXECD_1_1 = "http://tail-f.com/ns/execd/1.1"
#: Namespace for Cisco data model
CISCO_CPI_1_0 = "http://www.cisco.com/cpi_10/schema"
#: Namespace for Flowmon data model
FLOWMON_1_0 = "http://www.liberouter.org/ns/netopeer/flowmon/1.0"
#: Namespace for Juniper 9.6R4. Tested with Junos 9.6R4+
JUNIPER_1_1 = "http://xml.juniper.net/xnm/1.1/xnm"
#: Namespace for Huawei data model
HUAWEI_NS = "http://www.huawei.com/netconf/vrp"
#: Namespace for Huawei private
HW_PRIVATE_NS = "http://www.huawei.com/netconf/capability/base/1.0"
#: Namespace for H3C data model
H3C_DATA_1_0 = "http://www.h3c.com/netconf/data:1.0"
#: Namespace for H3C config model
H3C_CONFIG_1_0 = "http://www.h3c.com/netconf/config:1.0"
#: Namespace for H3C action model
H3C_ACTION_1_0 = "http://www.h3c.com/netconf/action:1.0"
#: Namespace for netconf monitoring
NETCONF_MONITORING_NS = "urn:ietf:params:xml:ns:yang:ietf-netconf-monitoring"
#: Namespace for netconf notifications
NETCONF_NOTIFICATION_NS = "urn:ietf:params:xml:ns:netconf:notification:1.0"
#: Namespace for netconf with-defaults (RFC 6243)
NETCONF_WITH_DEFAULTS_NS = "urn:ietf:params:xml:ns:yang:ietf-netconf-with-defaults"
#: Namespace for Alcatel-Lucent SR OS Base r13 YANG models
ALU_CONFIG = "urn:alcatel-lucent.com:sros:ns:yang:conf-r13"
#: Namespace for Nokia SR OS global operations
SROS_GLOBAL_OPS_NS = "urn:nokia.com:sros:ns:yang:sr:oper-global"


try:
    register_namespace = etree.register_namespace
except AttributeError:
    def register_namespace(prefix, uri):
        from xml.etree import ElementTree
        # cElementTree uses ElementTree's _namespace_map, so that's ok
        ElementTree._namespace_map[uri] = prefix

for (ns, pre) in {
    BASE_NS_1_0: 'nc',
    NETCONF_MONITORING_NS: 'ncm',
    NXOS_1_0: 'nxos',
    NXOS_IF: 'if',
    TAILF_AAA_1_1: 'aaa',
    TAILF_EXECD_1_1: 'execd',
    CISCO_CPI_1_0: 'cpi',
    FLOWMON_1_0: 'fm',
    JUNIPER_1_1: 'junos',
}.items():
    register_namespace(pre, ns)

qualify = lambda tag, ns=BASE_NS_1_0: tag if ns is None else "{%s}%s" % (ns, tag)
"""Qualify a *tag* name with a *namespace*, in :mod:`~xml.etree.ElementTree` fashion i.e. *{namespace}tagname*."""


def to_xml(ele, encoding="UTF-8", pretty_print=False):
    "Convert and return the XML for an *ele* (:class:`~xml.etree.ElementTree.Element`) with specified *encoding*."
    xml = etree.tostring(ele, encoding=encoding, pretty_print=pretty_print)
    return xml.decode('UTF-8') if xml.startswith(b'<?xml') \
        else '<?xml version="1.0" encoding="%s"?>%s' % (encoding, xml.decode('UTF-8'))


def to_ele(x, huge_tree=False):
    """Convert and return the :class:`~xml.etree.ElementTree.Element` for the XML document *x*. If *x* is already an :class:`~xml.etree.ElementTree.Element` simply returns that.

    *huge_tree*: parse XML with very deep trees and very long text content
    """
    return x if etree.iselement(x) else etree.fromstring(x.encode('UTF-8'), parser=_get_parser(huge_tree))


def parse_root(raw):
    "Efficiently parses the root element of a *raw* XML document, returning a tuple of its qualified name and attribute dictionary."
    fp = BytesIO(raw.encode('UTF-8'))
    for event, element in etree.iterparse(fp, events=('start',)):
        return (element.tag, element.attrib)

def validated_element(x, tags=None, attrs=None):
    """Checks if the root element of an XML document or Element meets the supplied criteria.

    *tags* if specified is either a single allowable tag name or sequence of allowable alternatives

    *attrs* if specified is a sequence of required attributes, each of which may be a sequence of several allowable alternatives

    Raises :exc:`XMLError` if the requirements are not met.
    """
    ele = to_ele(x)
    if tags:
        if isinstance(tags, (str, bytes)):
            tags = [tags]
        if ele.tag not in tags:
            raise XMLError("Element [%s] does not meet requirement" % ele.tag)
    if attrs:
        for req in attrs:
            if isinstance(req, (str, bytes)): req = [req]
            for alt in req:
                if alt in ele.attrib:
                    break
            else:
                raise XMLError("Element [%s] does not have required attributes" % ele.tag)
    return ele

XPATH_NAMESPACES = {
    're':'http://exslt.org/regular-expressions'
}


class NCElement:
    def __init__(self, result, transform_reply, huge_tree=False):
        self.__result = result
        self.__transform_reply = transform_reply
        self.__huge_tree = huge_tree
        if isinstance(transform_reply, types.FunctionType):
            self.__doc = self.__transform_reply(result._root)
        else:
            self.__doc = self.remove_namespaces(self.__result)

    def xpath(self, expression, namespaces={}):
        """Perform XPath navigation on an object

        Args:
            expression: A string representing a compliant XPath
                expression.
            namespaces: A dict of caller supplied prefix/xmlns to
                append to the static dict of XPath namespaces.
        Returns:
            A list of 'lxml.etree._Element' should a match on the
            expression be successful.  Otherwise, an empty list will
            be returned to the caller.
        """
        self.__expression = expression
        self.__namespaces = XPATH_NAMESPACES
        self.__namespaces.update(namespaces)
        return self.__doc.xpath(self.__expression, namespaces=self.__namespaces)

    def find(self, expression):
        """return result for a call to lxml ElementPath find()"""
        self.__expression = expression
        return self.__doc.find(self.__expression)

    def findtext(self, expression):
        """return result for a call to lxml ElementPath findtext()"""
        self.__expression = expression
        return self.__doc.findtext(self.__expression)

    def findall(self, expression):
        """return result for a call to lxml ElementPath findall()"""
        self.__expression = expression
        return self.__doc.findall(self.__expression)

    def __str__(self):
        """syntactic sugar for str() - alias to tostring"""
        return self.tostring.decode('UTF-8')

    @property
    def tostring(self):
        """return a pretty-printed string output for rpc reply"""
        parser = etree.XMLParser(remove_blank_text=True, huge_tree=self.__huge_tree)
        outputtree = etree.XML(etree.tostring(self.__doc), parser)
        return etree.tostring(outputtree, pretty_print=True)

    @property
    def data_xml(self):
        """return an unmodified output for rpc reply"""
        return to_xml(self.__doc)

    def remove_namespaces(self, rpc_reply):
        """remove xmlns attributes from rpc reply"""
        self.__xslt=self.__transform_reply
        self.__parser = etree.XMLParser(remove_blank_text=True, huge_tree=self.__huge_tree)
        self.__xslt_doc = etree.parse(io.BytesIO(self.__xslt), self.__parser)
        self.__transform = etree.XSLT(self.__xslt_doc)
        self.__root = etree.fromstring(str(self.__transform(etree.parse(StringIO(str(rpc_reply)),
                                                                        parser=self.__parser))),
                                       parser=self.__parser)
        return self.__root

def parent_ns(node):
    if node.prefix:
        return node.nsmap[node.prefix]
    return None

def yang_action(name, attrs):
    """Instantiate a YANG action element

    Args:
        name: A string representing the first descendant name of the
            XML element for the YANG action.
        attrs: A dict of attributes to apply to the XML element
            (e.g. namespaces).
    Returns:
        A tuple of 'lxml.etree._Element' values.  The first value
        represents the top-level YANG action element and the second
        represents the caller supplied initial node.
    """
    node = new_ele('action', attrs={'xmlns': YANG_NS_1_0})
    return (node, sub_ele(node, name, attrs))


def replace_namespace(root, old_ns, new_ns):
    """
    Substitute old_ns with new_ns for all the xml elements including and below root
    :param root: top element (root for this change)
    :param old_ns: old namespace
    :param new_ns: new namespace
    :return:
    """
    for elem in root.getiterator():
        # Comments don't have a namespace
        if elem.tag is not etree.Comment:
            # handle tag
            qtag = etree.QName(elem)
            if qtag.namespace == old_ns:
                elem.tag = etree.QName(new_ns, qtag.localname)

            # handle attributes
            attribs_dict = elem.attrib
            for attr in attribs_dict.keys():
                qattr = etree.QName(attr)
                if qattr.namespace == old_ns:
                    attribs_dict[etree.QName(new_ns, qattr.localname)] = attribs_dict.pop(attr)


new_ele_nsmap = lambda tag, nsmap, attrs={}, **extra: etree.Element(qualify(tag), attrs, nsmap, **extra)

new_ele = lambda tag, attrs={}, **extra: etree.Element(qualify(tag), attrs, **extra)

new_ele_ns = lambda tag, ns, attrs={}, **extra: etree.Element(qualify(tag,ns), attrs, **extra)

sub_ele = lambda parent, tag, attrs={}, **extra: etree.SubElement(parent, qualify(tag, parent_ns(parent)), attrs, **extra)

sub_ele_ns = lambda parent, tag, ns, attrs={}, **extra: etree.SubElement(parent, qualify(tag, ns), attrs, **extra)



# Copyright 2012 Vaibhav Bajpai
# Copyright 2009 Shikhar Bhushan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import getpass
import os
import re
import sys
import socket
import threading
from binascii import hexlify
from io import BytesIO as StringIO

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from ncclient.capabilities import Capabilities
from ncclient.logging_ import SessionLoggerAdapter

import paramiko

from ncclient.transport.errors import AuthenticationError, SSHError, SSHUnknownHostError
from ncclient.transport.session import Session
from ncclient.transport.parser import DefaultXMLParser

import logging
logger = logging.getLogger("ncclient.transport.ssh")

PORT_NETCONF_DEFAULT = 830

BUF_SIZE = 4096

#
# Define delimiters for chunks and messages for netconf 1.1 chunk enoding.
# When matched:
#
# * result.group(0) will contain whole matched string
# * result.group(1) will contain the digit string for a chunk
# * result.group(2) will be defined if '##' found
#
RE_NC11_DELIM = re.compile(br'\n(?:#([0-9]+)|(##))\n')


def default_unknown_host_cb(host, fingerprint):
    """An unknown host callback returns `True` if it finds the key acceptable, and `False` if not.

    This default callback always returns `False`, which would lead to :meth:`connect` raising a :exc:`SSHUnknownHost` exception.

    Supply another valid callback if you need to verify the host key programmatically.

    *host* is the hostname that needs to be verified

    *fingerprint* is a hex string representing the host key fingerprint, colon-delimited e.g. `"4b:69:6c:72:6f:79:20:77:61:73:20:68:65:72:65:21"`
    """
    return False


def _colonify(fp):
    fp = fp.decode('UTF-8')
    finga = fp[:2]
    for idx in range(2, len(fp), 2):
        finga += ":" + fp[idx:idx+2]
    return finga


class SSHSession(Session):

    "Implements a :rfc:`4742` NETCONF session over SSH."

    def __init__(self, device_handler):
        capabilities = Capabilities(device_handler.get_capabilities())
        Session.__init__(self, capabilities)
        self._host = None
        self._host_keys = paramiko.HostKeys()
        self._transport = None
        self._connected = False
        self._channel = None
        self._channel_id = None
        self._channel_name = None
        self._buffer = StringIO()
        self._device_handler = device_handler
        self._message_list = []
        self._closing = threading.Event()
        self.parser = DefaultXMLParser(self)  # SAX or DOM parser

        self.logger = SessionLoggerAdapter(logger, {'session': self})

    def _dispatch_message(self, raw):
        # Provide basic response message
        self.logger.info("Received message from host")
        # Provide complete response from host at debug log level
        self.logger.debug("Received:\n%s", raw)
        return super(SSHSession, self)._dispatch_message(raw)

    def _parse(self):
        "Messages ae delimited by MSG_DELIM. The buffer could have grown by a maximum of BUF_SIZE bytes everytime this method is called. Retains state across method calls and if a byte has been read it will not be considered again."
        return self.parser._parse10()

    def load_known_hosts(self, filename=None):

        """Load host keys from an openssh :file:`known_hosts`-style file. Can
        be called multiple times.

        If *filename* is not specified, looks in the default locations i.e. :file:`~/.ssh/known_hosts` and :file:`~/ssh/known_hosts` for Windows.
        """

        if filename is None:
            filename = os.path.expanduser('~/.ssh/known_hosts')
            try:
                self._host_keys.load(filename)
            except IOError:
                # for windows
                filename = os.path.expanduser('~/ssh/known_hosts')
                try:
                    self._host_keys.load(filename)
                except IOError:
                    pass
        else:
            self._host_keys.load(filename)

    def close(self):
        self._closing.set()
        if self._transport.is_active():
            self._transport.close()

        # Wait for the transport thread to close.
        while self.is_alive() and (self is not threading.current_thread()):
            self.join(10)

        if self._channel:
            self._channel.close()
        self._channel = None
        self._connected = False

    # REMEMBER to update transport.rst if sig. changes, since it is hardcoded there
    def connect(
            self,
            host,
            port                = PORT_NETCONF_DEFAULT,
            timeout             = None,
            unknown_host_cb     = default_unknown_host_cb,
            username            = None,
            password            = None,
            key_filename        = None,
            allow_agent         = True,
            hostkey_verify      = True,
            hostkey_b64         = None,
            look_for_keys       = True,
            ssh_config          = None,
            sock_fd             = None,
            bind_addr           = None,
            sock                = None,
            keepalive           = None,
            environment         = None):

        """Connect via SSH and initialize the NETCONF session. First attempts the publickey authentication method and then password authentication.

        To disable attempting publickey authentication altogether, call with *allow_agent* and *look_for_keys* as `False`.

        *host* is the hostname or IP address to connect to

        *port* is by default 830 (PORT_NETCONF_DEFAULT), but some devices use the default SSH port of 22 so this may need to be specified

        *timeout* is an optional timeout for socket connect

        *unknown_host_cb* is called when the server host key is not recognized. It takes two arguments, the hostname and the fingerprint (see the signature of :func:`default_unknown_host_cb`)

        *username* is the username to use for SSH authentication

        *password* is the password used if using password authentication, or the passphrase to use for unlocking keys that require it

        *key_filename* is a filename where a the private key to be used can be found

        *allow_agent* enables querying SSH agent (if found) for keys

        *hostkey_verify* enables hostkey verification from ~/.ssh/known_hosts

        *hostkey_b64* only connect when server presents a public hostkey matching this (obtain from server /etc/ssh/ssh_host_*pub or ssh-keyscan)

        *look_for_keys* enables looking in the usual locations for ssh keys (e.g. :file:`~/.ssh/id_*`)

        *ssh_config* enables parsing of an OpenSSH configuration file, if set to its path, e.g. :file:`~/.ssh/config` or to True (in this case, use :file:`~/.ssh/config`).

        *sock_fd* is an already open socket which shall be used for this connection. Useful for NETCONF outbound ssh. Use host=None together with a valid sock_fd number

        *bind_addr* is a (local) source IP address to use, must be reachable from the remote device.
        
        *sock* is an already open Python socket to be used for this connection.

        *keepalive* Turn on/off keepalive packets (default is off). If this is set, after interval seconds without sending any data over the connection, a "keepalive" packet will be sent (and ignored by the remote host). This can be useful to keep connections alive over a NAT.

        *environment* a dictionary containing the name and respective values to set
        """
        if not (host or sock_fd or sock):
            raise SSHError("Missing host, socket or socket fd")

        self._host = host

        # Optionally, parse .ssh/config
        config = {}
        if ssh_config is True:
            ssh_config = "~/.ssh/config" if sys.platform != "win32" else "~/ssh/config"
        if ssh_config is not None:
            config = paramiko.SSHConfig()
            with open(os.path.expanduser(ssh_config)) as ssh_config_file_obj:
                config.parse(ssh_config_file_obj)

            # Save default Paramiko SSH port so it can be reverted
            paramiko_default_ssh_port = paramiko.config.SSH_PORT

            # Change the default SSH port to the port specified by the user so expand_variables
            # replaces %p with the passed in port rather than 22 (the defauld paramiko.config.SSH_PORT)

            paramiko.config.SSH_PORT = port

            config = config.lookup(host)

            # paramiko.config.SSHconfig::expand_variables is called by lookup so we can set the SSH port
            # back to the default
            paramiko.config.SSH_PORT = paramiko_default_ssh_port

            host = config.get("hostname", host)
            if username is None:
                username = config.get("user")
            if key_filename is None:
                key_filename = config.get("identityfile")
            if timeout is None:
                timeout = config.get("connecttimeout")
                if timeout:
                    timeout = int(timeout)

        if hostkey_verify:
            userknownhostsfile = config.get("userknownhostsfile")
            if userknownhostsfile:
                self.load_known_hosts(os.path.expanduser(userknownhostsfile))
            else:
                self.load_known_hosts()

        if username is None:
            username = getpass.getuser()

        if sock_fd is None and sock is None:
            proxycommand = config.get("proxycommand")
            if proxycommand:
                self.logger.debug("Configuring Proxy. %s", proxycommand)
                if not isinstance(proxycommand, str):
                  proxycommand = [os.path.expanduser(elem) for elem in proxycommand]
                else:
                  proxycommand = os.path.expanduser(proxycommand)
                sock = paramiko.proxy.ProxyCommand(proxycommand)
            else:
                for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM):
                    af, socktype, proto, canonname, sa = res
                    try:
                        sock = socket.socket(af, socktype, proto)
                        sock.settimeout(timeout)
                    except socket.error:
                        continue
                    try:
                        if bind_addr:
                            sock.bind((bind_addr, 0))
                        sock.connect(sa)
                    except socket.error:
                        sock.close()
                        continue
                    break
                else:
                    raise SSHError("Could not open socket to %s:%s" % (host, port))
        elif sock is None:
            sock = socket.fromfd(int(sock_fd), socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)

        self._transport = paramiko.Transport(sock)
        self._transport.set_log_channel(logger.name)
        if config.get("compression") == 'yes':
            self._transport.use_compression()

        if hostkey_b64:
            # If we need to connect with a specific hostkey, negotiate for only its type
            hostkey_obj = None
            for key_cls in [paramiko.Ed25519Key, paramiko.RSAKey, paramiko.ECDSAKey]:
                try:
                    hostkey_obj = key_cls(data=base64.b64decode(hostkey_b64))
                except paramiko.SSHException:
                    # Not a key of this type - try the next
                    pass
            if not hostkey_obj:
                # We've tried all known host key types and haven't found a suitable one to use - bail
                raise SSHError("Couldn't find suitable paramiko key class for host key %s" % hostkey_b64)
            self._transport._preferred_keys = [hostkey_obj.get_name()]
        elif self._host_keys:
            # Else set preferred host keys to those we possess for the host
            # (avoids situation where known_hosts contains a valid key for the host, but that key type is not selected during negotiation)
            known_host_keys_for_this_host = self._host_keys.lookup(host) or {}
            host_port = '[%s]:%s' % (host, port)
            known_host_keys_for_this_host.update(self._host_keys.lookup(host_port) or {})
            if known_host_keys_for_this_host:
                self._transport._preferred_keys = list(known_host_keys_for_this_host)

        # Connect
        try:
            self._transport.start_client()
        except paramiko.SSHException as e:
            raise SSHError('Negotiation failed: %s' % e)

        if hostkey_verify:
            server_key_obj = self._transport.get_remote_server_key()
            fingerprint = _colonify(hexlify(server_key_obj.get_fingerprint()))

            is_known_host = False

            # For looking up entries for nonstandard (22) ssh ports in known_hosts
            # we enclose host in brackets and append port number
            known_hosts_lookups = [host, '[%s]:%s' % (host, port)]

            if hostkey_b64:
                # If hostkey specified, remote host /must/ use that hostkey
                if(hostkey_obj.get_name() == server_key_obj.get_name() and hostkey_obj.asbytes() == server_key_obj.asbytes()):
                    is_known_host = True
            else:
                # Check known_hosts
                is_known_host = any(self._host_keys.check(lookup, server_key_obj) for lookup in known_hosts_lookups)

            if not is_known_host and not unknown_host_cb(host, fingerprint):
                raise SSHUnknownHostError(known_hosts_lookups[0], fingerprint)

        # Authenticating with our private key/identity
        if key_filename is None:
            key_filenames = []
        elif isinstance(key_filename, (str, bytes)):
            key_filenames = [key_filename]
        else:
            key_filenames = key_filename

        self._auth(username, password, key_filenames, allow_agent, look_for_keys)

        self._connected = True      # there was no error authenticating
        self._closing.clear()

        if keepalive:
            self._transport.set_keepalive(keepalive)

        # TODO: leopoul: Review, test, and if needed rewrite this part
        subsystem_names = self._device_handler.get_ssh_subsystem_names()
        for subname in subsystem_names:
            self._channel = self._transport.open_session()
            self._channel_id = self._channel.get_id()
            channel_name = "%s-subsystem-%s" % (subname, str(self._channel_id))
            self._channel.set_name(channel_name)
            if environment:
                try:
                    self._channel.update_environment(environment)
                except paramiko.SSHException as e:
                    self.logger.info("%s (environment update rejected)", e)
                    handle_exception = self._device_handler.handle_connection_exceptions(self)
                    # Ignore the exception, since we continue to try the different
                    # subsystem names until we find one that can connect.
                    # have to handle exception for each vendor here
                    if not handle_exception:
                        continue
            try:
                self._channel.invoke_subsystem(subname)
            except paramiko.SSHException as e:
                self.logger.info("%s (subsystem request rejected)", e)
                handle_exception = self._device_handler.handle_connection_exceptions(self)
                # Ignore the exception, since we continue to try the different
                # subsystem names until we find one that can connect.
                # have to handle exception for each vendor here
                if not handle_exception:
                    continue
            self._channel_name = self._channel.get_name()
            self._post_connect(timeout)
            # for further upcoming RPC responses, vendor can chose their
            # choice of parser. Say DOM or SAX
            self.parser = self._device_handler.get_xml_parser(self)
            return
        raise SSHError("Could not open connection, possibly due to unacceptable"
                       " SSH subsystem name.")

    def _auth(self, username, password, key_filenames, allow_agent,
              look_for_keys):
        saved_exception = None
        encoded_password = None if password is None else password.encode('utf-8')
        
        for key_filename in key_filenames:
            try:
                key = paramiko.PKey.from_path(key_filename, encoded_password)
                self.logger.debug("Trying key %s from %s",
                                  hexlify(key.get_fingerprint()),
                                  key_filename)
                self._transport.auth_publickey(username, key)
                return
            except Exception as e:
                saved_exception = e
                self.logger.debug(e)

        if allow_agent:
            # resequence keys from agent using private key names
            prepend_agent_keys=[]
            append_agent_keys=list(paramiko.Agent().get_keys())

            for key_filename in key_filenames:
                if key_filename.endswith("-cert.pub"):
                    pubkey_filename=key_filename[:-9]+".pub"
                elif key_filename.endswith(".pub"):
                    pubkey_filename=key_filename[:-4]+".pub"
                else:
                    pubkey_filename=key_filename+".pub"
                try:
                    file_key=paramiko.PublicBlob.from_file(pubkey_filename).key_blob
                except (FileNotFoundError, ValueError):
                    continue

                for idx, agent_key in enumerate(append_agent_keys):
                    if agent_key.asbytes() == file_key:
                        self.logger.debug("Prioritising SSH agent key found in %s",key_filename )
                        try:
                            append_agent_keys[idx].load_certificate(key_filename)
                        except (FileNotFoundError, ValueError):
                            continue

                        prepend_agent_keys.append(append_agent_keys.pop(idx))
                        break

            agent_keys=tuple(prepend_agent_keys+append_agent_keys)

            for key in agent_keys:
                try:
                    self.logger.debug("Trying SSH agent key %s",
                                      hexlify(key.get_fingerprint()))
                    self._transport.auth_publickey(username, key)
                    return
                except Exception as e:
                    saved_exception = e
                    self.logger.debug(e)

        keyfiles = []
        if look_for_keys:
            rsa_key = os.path.expanduser("~/.ssh/id_rsa")
            ecdsa_key = os.path.expanduser("~/.ssh/id_ecdsa")
            ed25519_key = os.path.expanduser("~/.ssh/id_ed25519")
            if os.path.isfile(rsa_key):
                keyfiles.append(rsa_key)
            if os.path.isfile(ecdsa_key):
                keyfiles.append(ecdsa_key)
            if os.path.isfile(ed25519_key):
                keyfiles.append(ed25519_key)
            # look in ~/ssh/ for windows users:
            rsa_key = os.path.expanduser("~/ssh/id_rsa")
            ecdsa_key = os.path.expanduser("~/ssh/id_ecdsa")
            ed25519_key = os.path.expanduser("~/ssh/id_ed25519")
            if os.path.isfile(rsa_key):
                keyfiles.append(rsa_key)
            if os.path.isfile(ecdsa_key):
                keyfiles.append(ecdsa_key)
            if os.path.isfile(ed25519_key):
                keyfiles.append(ed25519_key)

        for filename in keyfiles:
            try:
                key = paramiko.PKey.from_path(filename, encoded_password)
                self.logger.debug("Trying discovered key %s in %s",
                                  hexlify(key.get_fingerprint()), filename)
                self._transport.auth_publickey(username, key)
                return
            except Exception as e:
                saved_exception = e
                self.logger.debug(e)

        if password is not None:
            try:
                self._transport.auth_password(username, password)
                return
            except Exception as e:
                saved_exception = e
                self.logger.debug(e)

        if saved_exception is not None:
            # need pep-3134 to do this right
            raise AuthenticationError(repr(saved_exception))

        raise AuthenticationError("No authentication methods available")

    def _transport_read(self):
        return self._channel.recv(BUF_SIZE)

    def _transport_write(self, data):
        return self._channel.send(data)

    def _transport_register(self, selector, event):
        selector.register(self._channel, event)

    def _send_ready(self):
        return self._channel.send_ready()

    @property
    def host(self):
        """Host this session is connected to, or None if not connected."""
        if hasattr(self, '_host'):
            return self._host
        return None

    @property
    def transport(self):
        "Underlying `paramiko.Transport <http://www.lag.net/paramiko/docs/paramiko.Transport-class.html>`_ object. This makes it possible to call methods like :meth:`~paramiko.Transport.set_keepalive` on it."
        return self._transport



import unittest
try:
    from unittest.mock import patch, MagicMock  # Python 3.4 and later
    getattr(MagicMock, 'assert_called_once')  # Python 3.6 and later
except (ImportError, AttributeError):
    from mock import patch, MagicMock
from ncclient import manager
from ncclient.devices.junos import JunosDeviceHandler
import logging


class TestManager(unittest.TestCase):

    @patch('ncclient.transport.SSHSession')
    def test_ssh(self, mock_ssh):
        m = MagicMock()
        mock_ssh.return_value = m
        conn = self._mock_manager()
        m.connect.assert_called_once_with(host='10.10.10.10',
                                          port=22,
                                          username='user',
                                          password='password',
                                          hostkey_verify=False, allow_agent=False,
                                          timeout=3)
        self.assertEqual(conn._session, m)
        self.assertEqual(conn._timeout, 10)

    @patch('ncclient.manager.connect_ssh')
    def test_connect_ssh(self, mock_ssh):
        manager.connect(host='host')
        mock_ssh.assert_called_once_with(host='host')

    @patch('ncclient.transport.SSHSession.load_known_hosts')
    @patch('ncclient.transport.SSHSession.connect')
    def test_connect_ssh1(self, mock_ssh, mock_load_known_hosts):
        manager.connect(host='host')
        mock_ssh.assert_called_once_with(host='host')
        mock_load_known_hosts.assert_not_called()

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_connect_ssh2(self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = manager.connect_ssh(host='10.10.10.10',
                                    port=22,
                                    username='user',
                                    password='password',
                                    timeout=3,
                                    hostkey_verify=False,
                                    allow_agent=False,
                                    keepalive=10)
        self.assertEqual(mock_trans.called, 1)

    @patch('ncclient.transport.SSHSession.connect')
    @patch('ncclient.transport.SSHSession.transport')
    @patch('ncclient.transport.SSHSession.close')
    def test_connect_exception(self, mock_close, mock_transport, mock_ssh):
        mock_ssh.side_effect = Exception
        try:
            manager.connect(host='host')
        except Exception:
            Exception("connect occured exception")
        mock_ssh.assert_called_once_with(host='host')

    @patch('ncclient.transport.SSHSession.connect')
    @patch('ncclient.transport.SSHSession.take_notification')
    def test_manager_take_notification(self, mock_take_notification, mock_ssh):
        mock_take_notification.return_value = "test_take_notification"
        conn = self._mock_manager()
        ret = conn.take_notification()
        mock_take_notification.assert_called_once_with(True, None)
        self.assertEqual(ret, "test_take_notification")

    @patch('ncclient.transport.SSHSession.connect')
    @patch('ncclient.operations.retrieve.GetConfig._request')
    def test_manager_getattr(self, mock_request, mock_ssh):
        conn = self._mock_manager()
        conn.get_config("running")
        mock_ssh.assert_called_once_with(host='10.10.10.10',
                                         port=22,
                                         username='user',
                                         password='password',
                                         timeout=3,
                                         hostkey_verify=False,
                                         allow_agent=False)

    @patch('ncclient.transport.SSHSession.connect')
    @patch('ncclient.transport.Session.send')
    @patch('ncclient.operations.rpc.RPC._request')
    def test_manager_getattr2(self, mock_request, mock_send, mock_ssh):
        conn = self._mock_manager()
        conn.get_edit('config')
        mock_ssh.assert_called_once_with(host='10.10.10.10',
                                         port=22,
                                         username='user',
                                         password='password',
                                         timeout=3,
                                         hostkey_verify=False,
                                         allow_agent=False)

    @patch('ncclient.manager.connect_ssh')
    def test_connect_ssh_with_hostkey_ed25519(self, mock_ssh):
        hostkey = 'AAAAC3NzaC1lZDI1NTE5AAAAIIiHpGSf8fla6tCwLpwshvMGmUK+B/0v5CsRu+5v4uT7'
        manager.connect(host='host', hostkey=hostkey)
        mock_ssh.assert_called_once_with(host='host', hostkey=hostkey)

    @patch('ncclient.manager.connect_ssh')
    def test_connect_ssh_with_hostkey_ecdsa(self, mock_ssh):
        hostkey = 'AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBFJV9xLkuntH3Ry0GmK4FjYlW+01Ik4j/gbW+i3yIx+YEkF0B3iM7kiyDPqvmOPuVGfW+gq5oQzzdvHKspNkw70='
        manager.connect(host='host', hostkey=hostkey)
        mock_ssh.assert_called_once_with(host='host', hostkey=hostkey)

    @patch('ncclient.manager.connect_ssh')
    def test_connect_ssh_with_hostkey_rsa(self, mock_ssh):
        hostkey = 'AAAAB3NzaC1yc2EAAAADAQABAAABAQDfEAdDrz3l8+PF510ivzWyX/pjpn3Cp6UgjJOinXz82e1LTURZhKwm8blcP8aWe8Uri65Roe6Q/H1WMaR3jFJj4UW2EZY5N+M4esPhoP/APOnDu2XNKy9AK9yD/Bu64TYgkIPQ/6FHdotcQdYTAJ+ac+YfJMp5mhVPnRIh4rlF08a0/tDHzLJVMEoXzp5nfVHcA4W3+5RRhklbct10U0jxHmG8Db9XbKiEbhWs/UMy59UpJ+zr7zLUYPRntgqqkpCyyfeHFNK1P6m3FmyT06QekOioCFmY05y65dkjAwBlaO1RKj1X1lgCirRWu4vxYBo9ewIGPZtuzeyp7jnl7kGV'
        manager.connect(host='host', hostkey=hostkey)
        mock_ssh.assert_called_once_with(host='host', hostkey=hostkey)

    @patch('ncclient.manager.connect_ssh')
    def test_connect_outbound_ssh(self, mock_ssh):
        manager.connect(host=None, sock_fd=6)
        mock_ssh.assert_called_once_with(host=None, sock_fd=6)

    @patch('ncclient.manager.connect_ioproc')
    def test_connect_ioproc(self, mock_ssh):
        manager.connect(host='localhost', device_params={'name': 'junos',
                                                        'local': True})
        mock_ssh.assert_called_once_with(host='localhost',
                            device_params={'local': True, 'name': 'junos'})

    @patch('paramiko.proxy.ProxyCommand')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_connect_with_ssh_config(self, mock_session, mock_hex, mock_trans, mock_proxy):
        log = logging.getLogger('TestManager.test_connect_with_ssh_config')
        ssh_config_path = 'test/unit/ssh_config'

        conn = manager.connect(host='fake_host',
                                    port=830,
                                    username='user',
                                    password='password',
                                    hostkey_verify=False,
                                    allow_agent=False,
                                    ssh_config=ssh_config_path)

        log.debug(mock_proxy.call_args[0][0])
        self.assertEqual(mock_proxy.called, 1)
        mock_proxy.assert_called_with('ssh -W 10.0.0.1:830 jumphost.domain.com')

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_ssh2(self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(mock_trans.called, 1)
        self.assertEqual(conn._timeout, 10)
        self.assertEqual(conn._device_handler.device_params, {'name': 'junos'})
        self.assertEqual(
            conn._device_handler.__class__.__name__,
            "JunosDeviceHandler")

    @patch('ncclient.transport.ssh.Session._post_connect')
    @patch('ncclient.transport.third_party.junos.ioproc.IOProc.connect')
    def test_ioproc(self, mock_connect, mock_ioproc):
        conn = manager.connect(host='localhost',
                                    port=22,
                                    username='user',
                                    password='password',
                                    timeout=3,
                                    hostkey_verify=False,
                                    device_params={'local': True, 'name': 'junos'},
                                    manager_params={'timeout': 10})
        self.assertEqual(mock_connect.called, 1)
        self.assertEqual(conn._timeout, 10)
        self.assertEqual(conn._device_handler.device_params, {'local': True, 'name': 'junos'})
        self.assertEqual(
            conn._device_handler.__class__.__name__,
            "JunosDeviceHandler")

    def test_make_device_handler(self):
        device_handler = manager.make_device_handler({'name': 'junos'})
        self.assertEqual(
            device_handler.__class__.__name__,
            "JunosDeviceHandler")

    def test_make_device_handler_provided_handler(self):
        device_handler = manager.make_device_handler(
            {'handler': JunosDeviceHandler})
        self.assertEqual(
            device_handler.__class__.__name__,
            "JunosDeviceHandler")

    @patch('ncclient.operations.LockContext')
    def test_manager_locked(self, mock_lock):
        conn = manager.Manager(None, None, timeout=20)
        conn.locked(None)
        mock_lock.assert_called_once_with(None, None, None)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_client_capability(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(
            conn.client_capabilities,
            conn._session.client_capabilities)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_custom_client_capability(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        custom_capabilities = [
                'urn:custom:capability:1.0',
                'urn:custom:capability:2.0'
        ]
        conn = self._mock_manager(nc_params={'capabilities': custom_capabilities})
        self.assertEqual(
            conn.client_capabilities,
            conn._session.client_capabilities)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_server_capability(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(
            conn.server_capabilities,
            conn._session.server_capabilities)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_channel_id(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(conn.channel_id, conn._session._channel_id)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_channel_name(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(conn.channel_name, conn._session._channel_name)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_channel_session_id(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(conn.session_id, conn._session.id)

    @patch('socket.socket')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_connected(
            self, mock_session, mock_hex, mock_trans, mock_socket):
        conn = self._mock_manager()
        self.assertEqual(conn.connected, True)

    @patch('ncclient.manager.Manager.HUGE_TREE_DEFAULT')
    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.rpc.RPC')
    def test_manager_huge_node(self, mock_rpc, mock_session, default_value):

        # Set default value to True only in this test through the default_value mock
        default_value = True

        # true should propagate all the way to the RPC
        conn = self._mock_manager()
        self.assertTrue(conn.huge_tree)
        conn.execute(mock_rpc)
        mock_rpc.assert_called_once()
        self.assertTrue(mock_rpc.call_args[1]['huge_tree'])

        # false should propagate all the way to the RPC
        conn.huge_tree = False
        self.assertFalse(conn.huge_tree)
        mock_rpc.reset_mock()
        conn.execute(mock_rpc)
        mock_rpc.assert_called_once()
        self.assertFalse(mock_rpc.call_args[1]['huge_tree'])

    def _mock_manager(self, nc_params={}):
        conn = manager.connect(host='10.10.10.10',
                                    port=22,
                                    username='user',
                                    password='password',
                                    timeout=3,
                                    hostkey_verify=False, allow_agent=False,
                                    device_params={'name': 'junos'},
                                    manager_params={'timeout': 10},
                                    nc_params=nc_params)
        return conn

    @patch('socket.socket')
    @patch('paramiko.Transport.start_client')
    @patch('paramiko.Transport.get_remote_server_key')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.SSHSession._auth')
    @patch('paramiko.Transport.open_session')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_manager_environment(
             self, mock_session_post_connect, mock_transport_open_session,
             mock_ssh_session_auth, mock_hex,
             mock_transport_get_remote_server_key, mock_transport_start_client,
             mock_socket):
        m = MagicMock()
        mock_transport_open_session.return_value = m
        env={"VAR1":"VALUE1"}
        conn = manager.connect(host='10.10.10.10',
                                    hostkey_verify=False, allow_agent=False,
                                    environment=env)
        self.assertEqual(conn.connected, True)
        m.update_environment.assert_called_once_with(env)

    @patch('socket.fromfd')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.Session._post_connect')
    def test_outbound_manager_connected(
            self, mock_session, mock_hex, mock_trans, mock_fromfd):
        conn = self._mock_outbound_manager()
        self.assertEqual(conn.connected, True)

    def _mock_outbound_manager(self):
        conn = manager.connect(host=None,
                                    sock_fd=6,
                                    username='user',
                                    password='password',
                                    device_params={'name': 'junos'},
                                    hostkey_verify=False, allow_agent=False)
        return conn

    @patch('socket.socket')
    @patch('ncclient.manager.connect_ssh')
    def test_call_home(self, mock_ssh, mock_socket_open):
        mock_connected_socket = MagicMock()
        mock_server_socket = MagicMock()
        mock_socket_open.return_value = mock_server_socket
        mock_server_socket.accept.return_value = (mock_connected_socket,
                                                  'remote.host')

        with manager.call_home(host='0.0.0.0', port=1234) as chm:
            mock_ssh.assert_called_once_with(host='0.0.0.0',
                                                port=1234,
                                                sock=mock_connected_socket)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestManager)
    unittest.TextTestRunner(verbosity=2).run(suite)



import unittest
from xml.etree import ElementTree
from ncclient.operations.util import *
try:
    from unittest.mock import MagicMock  # Python 3.4 and later
except ImportError:
    from mock import MagicMock

xml = """<filter type="xpath">
        <configuration>
            <system>
                <services/>
            </system>
        </configuration>
    </filter>"""


class TestUtils(unittest.TestCase):

    def test_one_of_1(self):
        self.assertEqual(one_of(None, 10, None), None)

    def test_one_of_2(self):
        self.assertRaises(OperationError,
            one_of, None, 10, 10)

    def test_one_of_3(self):
        self.assertRaises(OperationError,
            one_of, None, None)

    def test_datastore_url(self):
        node = new_ele("target")
        sub_ele(node, "candidate")
        result = ElementTree.tostring(
            datastore_or_url(
                "target",
                "candidate"))
        self.assertEqual(result, ElementTree.tostring(node))

    def test_datastore_url_2(self):
        node = new_ele("web")
        sub_ele(node, "url").text = "http://juniper.net"
        result = ElementTree.tostring(
            datastore_or_url(
                "web",
                "http://juniper.net",
                capcheck=MagicMock()))
        self.assertEqual(result, ElementTree.tostring(node))

    def test_datastore_url_3(self):
        node = new_ele("web")
        result = ElementTree.tostring(
            datastore_or_url(
                "web",
                "http://juniper.net"))
        self.assertEqual(result, ElementTree.tostring(node))

    def test_build_filter(self):
        reply = build_filter(xml)
        call = ElementTree.tostring(reply)
        self.assertEqual(call, ElementTree.tostring(to_ele(xml)))

    def test_build_filter_xpath(self):
        criteria = "configuration/system"
        filter = ("xpath", criteria)
        reply = build_filter(filter)
        call = ElementTree.tostring(reply)
        node = new_ele("filter", type="xpath")
        node.attrib["select"] = criteria
        self.assertEqual(call, ElementTree.tostring(node))

    def test_build_filter_xpath_ns(self):
        select = "configuration/system"
        ns = {"ns0": "http://www.xxx.org"}
        criteria = (ns, select)
        filter = ("xpath", criteria)
        reply = build_filter(filter)
        call = ElementTree.tostring(reply)
        node = new_ele_nsmap("filter", ns, type="xpath")
        node.attrib["select"] = select
        self.assertEqual(call, ElementTree.tostring(node))

    def test_build_filter_subtree(self):
        criteria =  """<configuration>
            <system>
                <services/>
            </system>
        </configuration>"""
        filter = ("subtree", criteria)
        reply = build_filter(filter, capcheck="cap")
        call = ElementTree.tostring(reply)
        node = new_ele("filter", type="subtree")
        node.append(to_ele(criteria))
        self.assertEqual(call, ElementTree.tostring(node))

    def test_build_filter_other(self):
        criteria =  """<configuration>
            <system>
                <services/>
            </system>
        </configuration>"""
        filter = ("text", criteria)
        self.assertRaises(OperationError,
            build_filter, filter, capcheck="cap")



# Copyright 2009 Shikhar Bhushan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Event, Lock, RLock
from uuid import uuid4

from ncclient.xml_ import *
from ncclient.logging_ import SessionLoggerAdapter
from ncclient.transport import SessionListener
from ncclient.operations import util
from ncclient.operations.errors import OperationError, TimeoutExpiredError, MissingCapabilityError

import logging
logger = logging.getLogger("ncclient.operations.rpc")


class RPCError(OperationError):

    "Represents an `rpc-error`. It is a type of :exc:`OperationError` and can be raised as such."

    tag_to_attr = {
        qualify("error-type"): "_type",
        qualify("error-tag"): "_tag",
        qualify("error-app-tag"): "_app_tag",
        qualify("error-severity"): "_severity",
        qualify("error-info"): "_info",
        qualify("error-path"): "_path",
        qualify("error-message"): "_message"
    }

    def __init__(self, raw, errs=None):
        self._raw = raw
        if errs is None:
            # Single RPCError
            self._errlist = None
            for attr in RPCError.tag_to_attr.values():
                setattr(self, attr, None)
            for subele in raw:
                attr = RPCError.tag_to_attr.get(subele.tag, None)
                if attr is not None:
                    setattr(self, attr, subele.text if attr != "_info" else to_xml(subele) )
            if self.message is not None:
                OperationError.__init__(self, self.message)
            else:
                OperationError.__init__(self, self.to_dict())
        else:
            # Multiple errors returned. Errors is a list of RPCError objs
            self._errlist = errs
            errlist = []
            for err in errs:
                if err.severity:
                    errsev = err.severity
                else:
                    errsev = 'undefined'
                if err.message:
                    errmsg = err.message
                else:
                    errmsg = 'not an error message in the reply. Enable debug'
                errordict = {"severity": errsev, "message":errmsg}
                errlist.append(errordict)
            # We are interested in the severity and the message
            self._severity = 'warning'
            self._message = "\n".join(["%s: %s" %(err['severity'].strip(), err['message'].strip()) for err in errlist])
            self.errors = errs
            has_error = filter(lambda higherr: higherr['severity'] == 'error', errlist)
            if has_error:
                self._severity = 'error'
            OperationError.__init__(self, self.message)

    def to_dict(self):
        return dict([ (attr[1:], getattr(self, attr)) for attr in RPCError.tag_to_attr.values() ])

    @property
    def xml(self):
        "The `rpc-error` element as returned in XML. \
        Multiple errors are returned as list of RPC errors"
        return self._raw

    @property
    def type(self):
        "The contents of the `error-type` element."
        return self._type

    @property
    def tag(self):
        "The contents of the `error-tag` element."
        return self._tag

    @property
    def app_tag(self):
        "The contents of the `error-app-tag` element."
        return self._app_tag

    @property
    def severity(self):
        "The contents of the `error-severity` element."
        return self._severity

    @property
    def path(self):
        "The contents of the `error-path` element if present or `None`."
        return self._path

    @property
    def message(self):
        "The contents of the `error-message` element if present or `None`."
        return self._message

    @property
    def info(self):
        "XML string or `None`; representing the `error-info` element."
        return self._info

    @property
    def errlist(self):
        "List of errors if this represents multiple errors, otherwise None."
        return self._errlist


class RPCReply:

    """Represents an *rpc-reply*. Only concerns itself with whether the operation was successful.

    *raw*: the raw unparsed reply

    *huge_tree*: parse XML with very deep trees and very long text content

    .. note::
        If the reply has not yet been parsed there is an implicit, one-time parsing overhead to
        accessing some of the attributes defined by this class.
    """

    ERROR_CLS = RPCError
    "Subclasses can specify a different error class, but it should be a subclass of `RPCError`."

    def __init__(self, raw, huge_tree=False, parsing_error_transform=None):
        self._raw = raw
        self._parsing_error_transform = parsing_error_transform
        self._parsed = False
        self._root = None
        self._errors = []
        self._huge_tree = huge_tree

    def __repr__(self):
        return self._raw

    def parse(self):
        "Parses the *rpc-reply*."
        if self._parsed: return
        root = self._root = to_ele(self._raw, huge_tree=self._huge_tree) # The <rpc-reply> element
        # Per RFC 4741 an <ok/> tag is sent when there are no errors or warnings
        ok = root.find(qualify("ok"))
        if ok is None:
            # Create RPCError objects from <rpc-error> elements
            error = root.find('.//'+qualify('rpc-error'))
            if error is not None:
                for err in root.getiterator(error.tag):
                    # Process a particular <rpc-error>
                    self._errors.append(self.ERROR_CLS(err))
        try:
            self._parsing_hook(root)
        except Exception as e:
            if self._parsing_error_transform is None:
                # re-raise as we have no workaround
                raise e

            # Apply device specific workaround and try again
            self._parsing_error_transform(root)
            self._parsing_hook(root)

        self._parsed = True

    def _parsing_hook(self, root):
        "No-op by default. Gets passed the *root* element for the reply."
        pass

    def set_parsing_error_transform(self, transform_function):
        self._parsing_error_transform = transform_function

    @property
    def xml(self):
        "*rpc-reply* element as returned."
        return self._raw

    @property
    def ok(self):
        "Boolean value indicating if there were no errors."
        self.parse()
        return not self.errors # empty list => false

    @property
    def error(self):
        "Returns the first :class:`RPCError` and `None` if there were no errors."
        self.parse()
        if self._errors:
            return self._errors[0]
        else:
            return None

    @property
    def errors(self):
        "List of `RPCError` objects. Will be empty if there were no *rpc-error* elements in reply."
        self.parse()
        return self._errors


class RPCReplyListener(SessionListener): # internal use

    # Use a re-entrant lock so nested/recursive attempts to create the listener
    # (e.g. during teardown while another RPC is in flight) do not deadlock.
    creation_lock = RLock()

    # one instance per session -- maybe there is a better way??
    def __new__(cls, session, device_handler):
        with RPCReplyListener.creation_lock:
            instance = session.get_listener_instance(cls)
            if instance is None:
                instance = object.__new__(cls)
                instance._lock = Lock()
                instance._id2rpc = {}
                instance._device_handler = device_handler
                #instance._pipelined = session.can_pipeline
                session.add_listener(instance)
                instance.logger = SessionLoggerAdapter(logger,
                                                       {'session': session})
            return instance

    def register(self, id, rpc):
        with self._lock:
            self._id2rpc[id] = rpc

    def callback(self, root, raw):
        tag, attrs = root
        if self._device_handler.perform_qualify_check():
            if tag != qualify("rpc-reply"):
                return
        if "message-id" not in attrs:
            # required attribute so raise OperationError
            raise OperationError("Could not find 'message-id' attribute in <rpc-reply>")
        else:
            id = attrs["message-id"]  # get the msgid
            with self._lock:
                try:
                    rpc = self._id2rpc[id]  # the corresponding rpc
                    self.logger.debug("Delivering to %r", rpc)
                    rpc.deliver_reply(raw)
                except KeyError:
                    raise OperationError("Unknown 'message-id': %s" % id)
                # no catching other exceptions, fail loudly if must
                else:
                    # if no error delivering, can del the reference to the RPC
                    del self._id2rpc[id]

    def errback(self, err):
        try:
            for rpc in self._id2rpc.values():
                rpc.deliver_error(err)
        finally:
            self._id2rpc.clear()


class RaiseMode:
    """
    Define how errors indicated by RPC should be handled.

    Note that any error_filters defined in the device handler will still be
    applied, even if ERRORS or ALL is defined: If the filter matches, an exception
    will NOT be raised.

    """
    NONE = 0
    "Don't attempt to raise any type of `rpc-error` as :exc:`RPCError`."

    ERRORS = 1
    "Raise only when the `error-type` indicates it is an honest-to-god error."

    ALL = 2
    "Don't look at the `error-type`, always raise."


class RPC:

    """Base class for all operations, directly corresponding to *rpc* requests. Handles making the request, and taking delivery of the reply."""

    DEPENDS = []
    """Subclasses can specify their dependencies on capabilities as a list of URI's or abbreviated names, e.g. ':writable-running'. These are verified at the time of instantiation. If the capability is not available, :exc:`MissingCapabilityError` is raised."""

    REPLY_CLS = RPCReply
    "By default :class:`RPCReply`. Subclasses can specify a :class:`RPCReply` subclass."


    def __init__(self, session, device_handler, async_mode=False, timeout=30, raise_mode=RaiseMode.NONE, huge_tree=False):
        """
        *session* is the :class:`~ncclient.transport.Session` instance

        *device_handler" is the :class:`~ncclient.devices.*.*DeviceHandler` instance

        *async* specifies whether the request is to be made asynchronously, see :attr:`is_async`

        *timeout* is the timeout for a synchronous request, see :attr:`timeout`

        *raise_mode* specifies the exception raising mode, see :attr:`raise_mode`

        *huge_tree* parse xml with huge_tree support (e.g. for large text config retrieval), see :attr:`huge_tree`
        """
        self._session = session
        try:
            for cap in self.DEPENDS:
                self._assert(cap)
        except AttributeError:
            pass
        self._async = async_mode
        self._timeout = timeout
        self._raise_mode = raise_mode
        self._huge_tree = huge_tree
        self._id = uuid4().urn # Keeps things simple instead of having a class attr with running ID that has to be locked
        self._listener = RPCReplyListener(session, device_handler)
        self._listener.register(self._id, self)
        self._reply = None
        self._error = None
        self._event = Event()
        self._device_handler = device_handler
        self.logger = SessionLoggerAdapter(logger, {'session': session})


    def _wrap(self, subele):
        # internal use
        ele = new_ele("rpc", {"message-id": self._id},
                      **self._device_handler.get_xml_extra_prefix_kwargs())
        ele.append(subele)
        return to_xml(ele)

    def _request(self, op):
        """Implementations of :meth:`request` call this method to send the request and process the reply.

        In synchronous mode, blocks until the reply is received and returns :class:`RPCReply`. Depending on the :attr:`raise_mode` a `rpc-error` element in the reply may lead to an :exc:`RPCError` exception.

        In asynchronous mode, returns immediately, returning `self`. The :attr:`event` attribute will be set when the reply has been received (see :attr:`reply`) or an error occured (see :attr:`error`).

        *op* is the operation to be requested as an :class:`~xml.etree.ElementTree.Element`
        """
        self.logger.info('Requesting %r', self.__class__.__name__)
        req = self._wrap(op)
        self._session.send(req)
        if self._async:
            self.logger.debug('Async request, returning %r', self)
            return self
        else:
            self.logger.debug('Sync request, will wait for timeout=%r', self._timeout)
            self._event.wait(self._timeout)
            if self._event.is_set():
                if self._error:
                    # Error that prevented reply delivery
                    raise self._error
                self._reply.parse()
                if self._reply.error is not None and not self._device_handler.is_rpc_error_exempt(self._reply.error.message):
                    # <rpc-error>'s [ RPCError ]

                    if self._raise_mode == RaiseMode.ALL or (self._raise_mode == RaiseMode.ERRORS and self._reply.error.severity == "error"):
                        errlist = []
                        errors = self._reply.errors
                        if len(errors) > 1:
                            raise RPCError(to_ele(self._reply._raw), errs=errors)
                        else:
                            raise self._reply.error
                if self._device_handler.transform_reply():
                    return NCElement(self._reply, self._device_handler.transform_reply(), huge_tree=self._huge_tree)
                else:
                    return self._reply
            else:
                raise TimeoutExpiredError('ncclient timed out while waiting for an rpc reply.')

    def request(self):
        """Subclasses must implement this method. Typically only the request needs to be built as an
        :class:`~xml.etree.ElementTree.Element` and everything else can be handed off to
        :meth:`_request`."""
        pass

    def _assert(self, capability):
        """Subclasses can use this method to verify that a capability is available with the NETCONF
        server, before making a request that requires it. A :exc:`MissingCapabilityError` will be
        raised if the capability is not available."""
        if capability not in self._session.server_capabilities:
            raise MissingCapabilityError('Server does not support [%s]' % capability)

    def deliver_reply(self, raw):
        # internal use
        self._reply = self.REPLY_CLS(raw, huge_tree=self._huge_tree)

        # Set the reply_parsing_error transform outside the constructor, to keep compatibility for
        # third party reply classes outside of ncclient
        self._reply.set_parsing_error_transform(
            self._device_handler.reply_parsing_error_transform(self.REPLY_CLS)
        )

        self._event.set()

    def deliver_error(self, err):
        # internal use
        self._error = err
        self._event.set()

    @property
    def reply(self):
        ":class:`RPCReply` element if reply has been received or `None`"
        return self._reply

    @property
    def error(self):
        """:exc:`Exception` type if an error occured or `None`.

        .. note::
            This represents an error which prevented a reply from being received. An *rpc-error*
            does not fall in that category -- see `RPCReply` for that.
        """
        return self._error

    @property
    def id(self):
        "The *message-id* for this RPC."
        return self._id

    @property
    def session(self):
        "The `~ncclient.transport.Session` object associated with this RPC."
        return self._session

    @property
    def event(self):
        """:class:`~threading.Event` that is set when reply has been received or when an error preventing
        delivery of the reply occurs.
        """
        return self._event

    def __set_async(self, async_mode=True):
        self._async = async_mode
        if async_mode and not self._session.can_pipeline:
            raise UserWarning('Asynchronous mode not supported for this device/session')

    def __set_raise_mode(self, mode):
        assert(mode in (RaiseMode.NONE, RaiseMode.ERRORS, RaiseMode.ALL))
        self._raise_mode = mode

    def __set_timeout(self, timeout):
        self._timeout = timeout

    raise_mode = property(fget=lambda self: self._raise_mode, fset=__set_raise_mode)
    """Depending on this exception raising mode, an `rpc-error` in the reply may be raised as an :exc:`RPCError` exception. Valid values are the constants defined in :class:`RaiseMode`. """

    is_async = property(fget=lambda self: self._async, fset=__set_async)
    """Specifies whether this RPC will be / was requested asynchronously. By default RPC's are synchronous."""

    timeout = property(fget=lambda self: self._timeout, fset=__set_timeout)
    """Timeout in seconds for synchronous waiting defining how long the RPC request will block on a reply before raising :exc:`TimeoutExpiredError`.

    Irrelevant for asynchronous usage.
    """

    @property
    def huge_tree(self):
        """Whether `huge_tree` support for XML parsing of RPC replies is enabled (default=False)"""
        return self._huge_tree

    @huge_tree.setter
    def huge_tree(self, x):
        self._huge_tree = x

class GenericRPC(RPC):
    """Generic rpc commands wrapper"""
    REPLY_CLS = RPCReply
    """See :class:`RPCReply`."""

    def request(self, rpc_command, source=None, filter=None, config=None, target=None, format=None):
        """
        *rpc_command* specifies rpc command to be dispatched either in plain text or in xml element format (depending on command)

        *target* name of the configuration datastore being edited

        *source* name of the configuration datastore being queried

        *config* is the configuration, which must be rooted in the `config` element. It can be specified either as a string or an :class:`~xml.etree.ElementTree.Element`.

        *filter* specifies the portion of the configuration to retrieve (by default entire configuration is retrieved)

        :seealso: :ref:`filter_params`

        Examples of usage::

            m.rpc('rpc_command')

        or dispatch element like ::

            rpc_command = new_ele('get-xnm-information')
            sub_ele(rpc_command, 'type').text = "xml-schema"
            m.rpc(rpc_command)
        """

        if etree.iselement(rpc_command):
            node = rpc_command
        else:
            node = new_ele(rpc_command)
        if target is not None:
            node.append(util.datastore_or_url("target", target, self._assert))
        if source is not None:
            node.append(util.datastore_or_url("source", source, self._assert))
        if filter is not None:
            node.append(util.build_filter(filter))
        if config is not None:
            node.append(validated_element(config, ("config", qualify("config"))))

        return self._request(node)



import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.huawei.rpc import *


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'huawei'})
     
    @patch('ncclient.operations.third_party.huawei.rpc.RPC._request')
    def test_CLI(self, mock_request):
        mock_request.return_value = 'huawei'
        expected = 'huawei'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = CLI(session, self.device_handler)
        command = '<get>devices-name</get>'
        actual = obj.request(command=command)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.huawei.rpc.RPC._request')
    def test_action(self, mock_request):
        mock_request.return_value = 'huawei'
        expected = 'huawei'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Action(session, self.device_handler)
        action = '<get>devices-name</get>'
        actual = obj.request(action=action)
        self.assertEqual(expected, actual)




import copy
import unittest
from unittest.mock import patch

from xml.etree import ElementTree
from lxml import etree

from ncclient.operations.retrieve import *
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.capabilities import Capabilities
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from ncclient.operations.errors import MissingCapabilityError


class TestRetrieve(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'junos'})

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)
        root_filter = new_ele('filter')
        config_filter = sub_ele(root_filter, 'configuration')
        system_filter = sub_ele(config_filter, 'system')
        sub_ele(system_filter, 'services')
        obj.request(copy.deepcopy(root_filter))
        node = new_ele("get")
        node.append(util.build_filter(root_filter))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_with_defaults_basic_mode(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([
            "urn:ietf:params:netconf:capability:with-defaults:1.0"
            "?basic-mode=explicit"
        ])
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(with_defaults='explicit')

        expected_xml = etree.fromstring(
            '<nc:get xmlns:nc="{base}">'
            '<ns0:with-defaults xmlns:ns0="{defaults}">explicit</ns0:with-defaults>'
            '</nc:get>'.format(
                base=BASE_NS_1_0,
                defaults=NETCONF_WITH_DEFAULTS_NS
            )
        )

        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(etree.tostring(call), etree.tostring(expected_xml))

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_with_defaults_also_supported(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([
            "urn:ietf:params:netconf:capability:with-defaults:1.0"
            "?basic-mode=explicit"
            "&also-supported=report-all,trim"
        ])
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(with_defaults='report-all')

        expected_xml = etree.fromstring(
            '<nc:get xmlns:nc="{base}">'
            '<ns0:with-defaults xmlns:ns0="{defaults}">report-all</ns0:with-defaults>'
            '</nc:get>'.format(
                base=BASE_NS_1_0,
                defaults=NETCONF_WITH_DEFAULTS_NS
            )
        )

        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(etree.tostring(call), etree.tostring(expected_xml))

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_with_defaults_not_supported(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([
            "urn:ietf:params:netconf:capability:with-defaults:1.0"
            "?basic-mode=explicit"
            "&also-supported=report-all,trim"
        ])
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)

        expected_error = (
            "Invalid 'with-defaults' mode 'report-all-tagged'; the server "
            "only supports the following: explicit, report-all, trim"
        )
        self.assertRaisesRegex(
            WithDefaultsError,
            expected_error,
            obj.request,
            with_defaults='report-all-tagged'
        )

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_with_defaults_missing_capability(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([])
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)
        self.assertRaises(
            MissingCapabilityError,
            obj.request,
            with_defaults='report-all'
        )

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_config(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = GetConfig(session, self.device_handler, raise_mode=RaiseMode.ALL)
        source = "candidate"
        obj.request(source)
        node = new_ele("get-config")
        node.append(util.datastore_or_url("source", source))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_config_with_defaults(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([
            "urn:ietf:params:netconf:capability:with-defaults:1.0"
            "?basic-mode=explicit"
        ])
        obj = GetConfig(session, self.device_handler, raise_mode=RaiseMode.ALL)
        source = 'candidate'
        obj.request(source, with_defaults='explicit')

        expected_xml = etree.fromstring(
            '<nc:get-config xmlns:nc="{base}">'
            '<nc:source><nc:candidate /></nc:source>'
            '<ns0:with-defaults xmlns:ns0="{defaults}">explicit</ns0:with-defaults>'
            '</nc:get-config>'.format(
                base=BASE_NS_1_0,
                defaults=NETCONF_WITH_DEFAULTS_NS
            )
        )

        call = mock_request.call_args_list[0][0][0]
        self.assertEqual(etree.tostring(call), etree.tostring(expected_xml))

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_config_with_defaults_missing_capability(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = Capabilities([])
        obj = GetConfig(session, self.device_handler, raise_mode=RaiseMode.ALL)
        source = 'candidate'
        self.assertRaises(
            MissingCapabilityError,
            obj.request,
            source,
            with_defaults='explicit'
        )

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_schema(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = GetSchema(session, self.device_handler, raise_mode=RaiseMode.ALL)
        identifier = "foo"
        version = "1.0"
        reqformat = "xsd"
        obj.request(identifier, version, reqformat)
        node = etree.Element(qualify("get-schema", NETCONF_MONITORING_NS))
        id = etree.SubElement(node,
                              qualify("identifier", NETCONF_MONITORING_NS))
        id.text = identifier
        ver = etree.SubElement(node, qualify("version", NETCONF_MONITORING_NS))
        ver.text = version
        formt = etree.SubElement(node,
                                 qualify("format", NETCONF_MONITORING_NS))
        formt.text = reqformat
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_dispatch(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Dispatch(session, self.device_handler, raise_mode=RaiseMode.ALL)
        rpc = 'get-software-information'
        source = "candidate"
        root_filter = new_ele('filter')
        config_filter = sub_ele(root_filter, 'configuration')
        system_filter = sub_ele(config_filter, 'system')
        sub_ele(system_filter, 'services')
        a = copy.deepcopy(root_filter)
        obj.request(rpc, source=source, filter=a)
        node = new_ele(rpc)
        node.append(util.datastore_or_url("source", source))
        node.append(util.build_filter(root_filter))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_dispatch_2(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Dispatch(session, self.device_handler, raise_mode=RaiseMode.ALL)
        node = new_ele('get-software-information')
        source = "candidate"
        root_filter = new_ele('filter')
        config_filter = sub_ele(root_filter, 'configuration')
        system_filter = sub_ele(config_filter, 'system')
        sub_ele(system_filter, 'services')
        a = copy.deepcopy(root_filter)
        obj.request(node, source=source, filter=a)
        node.append(util.datastore_or_url("source", source))
        node.append(util.build_filter(root_filter))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.retrieve.RPC._request')
    def test_get_with_multi_subtree_filters(self, mock_request):
        result = '''
                <?xml version="1.0" encoding="UTF-8"?>
                <data xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
                      xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
                    <cont1 xmlns="urn:mod1">
                        <le1>test_mod1_001</le1>
                        <le2>this is a test-one example</le2>
                    </cont1>
                    <cont2 xmlns="urn:mod2">
                        <le1>test_mod2_002</le1>
                        <le2>this is a test-two example</le2>
                        <lst>
                            <le3>a list of mod2</le3>
                        </lst>
                    </cont2>
                </data>'''
        mock_request.return_value = result
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = Get(session, self.device_handler, raise_mode=RaiseMode.ALL)

        multi_subtree_filters = [
            '<cont1 xmlns="urn:mod1"> \
                <le1/> \
                <le2/> \
             </cont1>',
             '<cont2 xmlns="urn:mod2"/>'
        ]

        ret = obj.request(copy.deepcopy(multi_subtree_filters))
        node = new_ele("get")
        node.append(util.build_filter(multi_subtree_filters))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)
        self.assertEqual(ret, result)




from ncclient.xml_ import *

from ncclient.operations.rpc import RPC
from ncclient.operations.rpc import RPCReply
from ncclient.operations.rpc import RPCError
from ncclient import NCClientError
import math


class GetConfiguration(RPC):
    def request(self, format='xml', filter=None):
        node = new_ele('get-configuration', {'format':format})
        if filter is not None:
            node.append(filter)
        if format !='xml':
            # The entire config comes as a single text element and this requires huge_tree support for large configs
            self._huge_tree = True
        return self._request(node)

class LoadConfiguration(RPC):
    def request(self, format='xml', action='merge',
            target='candidate', config=None):
        if config is not None:
            if type(config) == list:
                config = '\n'.join(config)
            if action == 'set':
                format = 'text'
            node = new_ele('load-configuration', {'action':action, 'format':format})
            if format == 'xml':
                config_node = sub_ele(node, 'configuration')
                config_node.append(config)
            if format == 'json':
                config_node = sub_ele(node, 'configuration-json').text = config
            if format == 'text' and not action == 'set':
                config_node = sub_ele(node, 'configuration-text').text = config
            if action == 'set' and format == 'text':
                config_node = sub_ele(node, 'configuration-set').text = config
            return self._request(node)


class CompareConfiguration(RPC):
    def request(self, rollback=0, format='text'):
        node = new_ele('get-configuration', {'compare':'rollback', 'format':format, 'rollback':str(rollback)})
        return self._request(node)


class ExecuteRpc(RPC):
    def request(self, rpc, filter_xml=None):
        if isinstance(rpc, str):
            rpc = to_ele(rpc)
        self._filter_xml = filter_xml
        return self._request(rpc)


class Command(RPC):
    def request(self, command=None, format='xml'):
        node = new_ele('command', {'format':format})
        node.text = command
        return self._request(node)


class Reboot(RPC):
    def request(self):
        node = new_ele('request-reboot')
        return self._request(node)


class Halt(RPC):
    def request(self):
        node = new_ele('request-halt')
        return self._request(node)


class Commit(RPC):
    "`commit` RPC. Depends on the `:candidate` capability, and the `:confirmed-commit`."

    DEPENDS = [':candidate']

    def request(self, confirmed=False, timeout=None, comment=None, synchronize=False, at_time=None, check=False):
        """Commit the candidate configuration as the device's new current configuration. Depends on the `:candidate` capability.

        A confirmed commit (i.e. if *confirmed* is `True`) is reverted if there is no followup commit within the *timeout* interval. If no timeout is specified the confirm timeout defaults to 600 seconds (10 minutes). A confirming commit may have the *confirmed* parameter but this is not required. Depends on the `:confirmed-commit` capability.

        *confirmed* whether this is a confirmed commit. Mutually exclusive with at_time.

        *timeout* specifies the confirm timeout in seconds

        *comment* a string to comment the commit with. Review on the device using 'show system commit'

        *synchronize* Whether we should synch this commit across both Routing Engines

        *at_time* Mutually exclusive with confirmed. The time at which the commit should happen. Junos expects either of these two formats:
            A time value of the form hh:mm[:ss] (hours, minutes, and, optionally, seconds)
            A date and time value of the form yyyy-mm-dd hh:mm[:ss] (year, month, date, hours, minutes, and, optionally, seconds)

        *check* Verify the syntactic correctness of the candidate configuration"""
        # NOTE: non netconf standard, Junos specific commit-configuration element, see
        # https://www.juniper.net/documentation/en_US/junos/topics/reference/tag-summary/junos-xml-protocol-commit-configuration.html
        node = new_ele_ns("commit-configuration", "")
        if confirmed and at_time is not None:
            raise NCClientError("'Commit confirmed' and 'commit at' are mutually exclusive.")
        if confirmed:
            self._assert(":confirmed-commit")
            sub_ele(node, "confirmed")
            if timeout is not None:
                # NOTE: For Junos, confirm-timeout has to be given in minutes:
                # https://www.juniper.net/documentation/en_US/junos/topics/reference/tag-summary/junos-xml-protocol-commit-configuration.html
                # but the netconf standard and ncclient library uses seconds:
                # https://tools.ietf.org/html/rfc6241#section-8.4.5
                # http://ncclient.readthedocs.io/en/latest/manager.html#ncclient.manager.Manager.commit
                # so need to convert the value in seconds to minutes.
                timeout_int = int(timeout) if isinstance(timeout, str) else timeout
                sub_ele(node, "confirm-timeout").text = str(int(math.ceil(timeout_int/60.0)))
        elif at_time is not None:
            sub_ele(node, "at-time").text = at_time
        if comment is not None:
            sub_ele(node, "log").text = comment
        if synchronize:
            sub_ele(node, "synchronize")
        if check:
            sub_ele(node, "check")
        return self._request(node)

class Rollback(RPC):
    def request(self, rollback=0):
        node = new_ele('load-configuration', {'rollback':str(rollback)})
        return self._request(node)



from .default import DefaultDeviceHandler
from ncclient.xml_ import BASE_NS_1_0


class CienaDeviceHandler(DefaultDeviceHandler):
    """
    Ciena handler for device specific information.
    """

    def __init__(self, device_params, ignore_errors=None):
        super(CienaDeviceHandler, self).__init__(device_params, ignore_errors)

    def get_xml_base_namespace_dict(self):
        return {None: BASE_NS_1_0}

    def get_xml_extra_prefix_kwargs(self):
        d = {}
        d.update(self.get_xml_base_namespace_dict())
        return {"nsmap": d}



import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.iosxe.rpc import *


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'iosxe'})
    
    @patch('ncclient.operations.third_party.iosxe.rpc.RPC._request')
    def test_saveConfig(self, mock_request):
        mock_request.return_value = 'iosxe'
        expected = 'iosxe'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = SaveConfig(session, self.device_handler)
        actual = obj.request()
        self.assertEqual(expected, actual)



import unittest
try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.alu.rpc import *


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'alu'})
    
    @patch('ncclient.operations.third_party.alu.rpc.RPC._request')
    def test_showCLI(self, mock_request):
        mock_request.return_value = 'alu'
        expected = 'alu'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = ShowCLI(session, self.device_handler)
        command = 'show system users'
        actual = obj.request(command=command)
        self.assertEqual(expected, actual)

    @patch('ncclient.operations.third_party.alu.rpc.RPC._request')
    def test_getConfiguration(self, mock_request):
        mock_request.return_value = 'alu'
        expected = 'alu'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = GetConfiguration(session, self.device_handler)
        content = 'xml'
        actual = obj.request(content=content)
        self.assertEqual(expected, actual)

        filter = '<get>device-name</get>'
        actual = obj.request(content=content, filter=filter)
        self.assertEqual(expected, actual)

        content = 'cli'
        actual = obj.request(content=content, filter=filter)
        self.assertEqual(expected, actual)
        
        detail = True
        actual = obj.request(content=content, filter=filter, detail=detail)
        self.assertEqual(expected, actual)
        
        content = ''
        actual = obj.request(content=content, filter=filter, detail=detail)
        self.assertEqual(expected, actual)
        
    @patch('ncclient.operations.third_party.alu.rpc.RPC._request')
    def test_loadConfiguration(self, mock_request):
        mock_request.return_value = 'alu'
        expected = 'alu'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = LoadConfiguration(session, self.device_handler)
        default_operation = ''
        format = 'xml'
        actual = obj.request(format=format, default_operation=default_operation)
        self.assertEqual(expected, actual)

        default_operation = 'get'
        actual=obj.request(format=format, default_operation=default_operation)
        self.assertEqual(expected, actual)

        config = new_ele('device-name')
        actual=obj.request(format=format, default_operation=default_operation, config=config)
        self.assertEqual(expected, actual)

        config = 'device-name'
        format = 'cli'
        actual=obj.request(format=format, default_operation=default_operation, config=config)
        self.assertEqual(expected, actual)
        
        default_operation = ''
        actual=obj.request(format=format, default_operation=default_operation, config=config)
        self.assertEqual(expected, actual)



import unittest
from ncclient.devices.h3c import *


capabilities = ['urn:ietf:params:netconf:base:1.0',
                'urn:ietf:params:netconf:base:1.1',
                'urn:ietf:params:netconf:capability:writable-running:1.0',
                'urn:ietf:params:netconf:capability:candidate:1.0',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
                'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
                'urn:ietf:params:netconf:capability:startup:1.0',
                'urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file,https,sftp',
                'urn:ietf:params:netconf:capability:validate:1.0',
                'urn:ietf:params:netconf:capability:xpath:1.0',
                'urn:ietf:params:netconf:capability:notification:1.0',
                'urn:ietf:params:netconf:capability:interleave:1.0',
                'urn:ietf:params:netconf:capability:with-defaults:1.0']


class TestH3cDevice(unittest.TestCase):

    def setUp(self):
        self.obj = H3cDeviceHandler({'name': 'h3c'})

    def test_add_additional_operations(self):
        expected = dict()
        expected['get_bulk'] = GetBulk
        expected['get_bulk_config'] = GetBulkConfig
        expected['cli'] = CLI
        expected['action'] = Action
        expected['save'] = Save
        expected['load'] = Load
        expected['rollback'] = Rollback
        self.assertDictEqual(expected, self.obj.add_additional_operations())

    def test_get_capabilities(self):
        self.assertListEqual(capabilities, self.obj.get_capabilities())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected['nsmap'] = self.obj.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())

    def test_perform_qualify_check(self):
        self.assertFalse(self.obj.perform_qualify_check())



# Copyright (c) Siemens AG, 2022
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl
import sys
import unittest
import socket
from unittest.mock import MagicMock, patch, call

try:
    from ssl import PROTOCOL_TLS_CLIENT
    tls_proto = PROTOCOL_TLS_CLIENT
except ImportError:
    from ssl import PROTOCOL_TLS
    tls_proto = PROTOCOL_TLS

from ncclient.transport.errors import TLSError
from ncclient.transport.tls import TLSSession, DEFAULT_TLS_NETCONF_PORT


HOST = '10.10.10.10'
PORT = DEFAULT_TLS_NETCONF_PORT
KEYFILE = 'test/unit/transport/certs/test.key'
CERTFILE_WO_KEY = 'test/unit/transport/certs/test.crt'
CERTFILE_WITH_KEY = 'test/unit/transport/certs/test.pem'


class TestTLS(unittest.TestCase):

    @patch('ssl.SSLContext.wrap_socket')
    @patch('ncclient.transport.Session._post_connect')
    def test_connect_tls_certfile_with_key(self, mock_post_connect, mock_wrap_socket_fn):
        session = TLSSession(MagicMock())
        mock_wrap_socket = MagicMock()
        mock_wrap_socket_fn.return_value = mock_wrap_socket
        session.connect(host=HOST, certfile=CERTFILE_WITH_KEY,
                        protocol=tls_proto)
        mock_wrap_socket.connect.assert_called_once_with(
            (HOST, DEFAULT_TLS_NETCONF_PORT))
        self.assertTrue(session.connected)

    @patch('ssl.SSLContext.wrap_socket')
    @patch('ncclient.transport.Session._post_connect')
    def test_connect_tls_certfile_without_key(self, mock_post_connect, mock_wrap_socket_fn):
        session = TLSSession(MagicMock())
        mock_wrap_socket = MagicMock()
        mock_wrap_socket_fn.return_value = mock_wrap_socket
        session.connect(host=HOST, certfile=CERTFILE_WO_KEY, keyfile=KEYFILE,
                        protocol=tls_proto)
        mock_wrap_socket.connect.assert_called_once_with(
            (HOST, DEFAULT_TLS_NETCONF_PORT))
        self.assertTrue(session.connected)

    @unittest.skipIf(sys.version_info < (3, 6), "test not supported < Python3.6")
    @patch('ssl.SSLContext.wrap_socket')
    @patch('ncclient.transport.Session._post_connect')
    def test_tls_handshake(self, mock_post_connect, mock_wrap_socket_fn):
        session = TLSSession(MagicMock())
        mock_wrap_socket = MagicMock()
        mock_wrap_socket_fn.return_value = mock_wrap_socket
        session.connect(host=HOST, certfile=CERTFILE_WITH_KEY,
                        protocol=tls_proto)
        mock_wrap_socket.do_handshake.assert_called_once()

    @patch('ssl.SSLContext.wrap_socket')
    def test_tls_unsuccessul_handshake(self, mock_wrap_socket_fn):
        session = TLSSession(MagicMock())
        mock_wrap_socket = MagicMock()
        mock_wrap_socket.do_handshake = MagicMock(side_effect=ssl.SSLError)
        mock_wrap_socket_fn.return_value = mock_wrap_socket
        with self.assertRaises(TLSError):
            session.connect(host=HOST, certfile=CERTFILE_WITH_KEY,
                            protocol=tls_proto)

    def test_connect_tls_missing_hostname(self):
        session = TLSSession(MagicMock())
        with self.assertRaises(TLSError):
            session.connect(host=None, certfile=CERTFILE_WITH_KEY,
                            protocol=tls_proto)

    def test_connect_tls_missing_certfile(self):
        session = TLSSession(MagicMock())
        with self.assertRaises(TLSError):
            session.connect(host=HOST, certfile=None,
                            protocol=tls_proto)

    def test_connect_tls_missing_private_key(self):
        session = TLSSession(MagicMock())
        with self.assertRaises(TLSError):
            session.connect(host=HOST, certfile=CERTFILE_WO_KEY,
                            protocol=tls_proto)

    def test_connect_tls_missing_ca_file(self):
        session = TLSSession(MagicMock())
        with self.assertRaises(TLSError):
            # Use arbitrary string to point to non-existing key location.
            session.connect(host=HOST, certfile=CERTFILE_WITH_KEY,
                            ca_certs='-', protocol=tls_proto)

    @patch('socket.socket.close')
    def test_close_tls(self, mock_sock_close_fn):
        session = TLSSession(MagicMock())
        session._socket = socket.socket()
        session._connected = True
        session.close()
        mock_sock_close_fn.assert_called_once_with()
        self.assertFalse(session._connected)



#!/usr/bin/env python
#
# For Nokia SR OS > 20.10, support for issuing a subset of CLI commands
# (e.g. 'show', 'admin', 'clear', 'ping', etc..) over NETCONF RPCs is
# achieved by the use of the <md-cli-raw-command> YANG 1.1 action

import logging
import sys

from ncclient import manager
from ncclient.operations.rpc import RPCError

_NS_MAP = {
    'nc': 'urn:ietf:params:xml:ns:netconf:base:1.0',
    'nokia-oper': 'urn:nokia.com:sros:ns:yang:sr:oper-global'
}

def connect(host, port, user, password):
    m = manager.connect(host=host, port=port,
            username=user, password=password,
            device_params={'name': 'sros'},
            hostkey_verify=False)

    ## Issue 'show card' and display the raw XML output
    result = m.md_cli_raw_command('show card')
    logging.info(result)

    ## Issue 'show port' from MD-CLI context and emit only
    ## the returned command contents
    result = m.md_cli_raw_command('show port')
    output = result.xpath(
            '/nc:rpc-reply/nokia-oper:results' \
            '/nokia-oper:md-cli-output-block',
            namespaces=_NS_MAP)[0].text
    logging.info(output)

    ## Issue 'show version' from Classic CLI context and emit
    ## only the returned command contents
    result = m.md_cli_raw_command('//show version')
    output = result.xpath(
            '/nc:rpc-reply/nokia-oper:results' \
            '/nokia-oper:md-cli-output-block',
            namespaces=_NS_MAP)[0].text
    logging.info(output)

    ## Issue 'ping' from MD-CLI context and emit only the
    ## returned command contents
    result = m.md_cli_raw_command(
            'ping 127.0.0.1 router-instance "Base" count 3')
    output = result.xpath(
            '/nc:rpc-reply/nokia-oper:results' \
            '/nokia-oper:md-cli-output-block',
            namespaces=_NS_MAP)[0].text
    logging.info(output)

    ## Issue an admin command that returns only NETCONF <ok/> or
    ## <rpc-error/>
    result = m.md_cli_raw_command('admin save')
    try:
        if len(result.xpath('/nc:rpc-reply/nc:ok', namespaces=_NS_MAP)) > 0:
            logging.info('Admin save successful')
        else:
            logging.info('Admin save unsuccessful')

    except RPCError as err:
        logging.info('Error: %s' % err.message.strip())


    ## Issue an unsupported command and handle the RPC error gracefully
    try:
        result = m.md_cli_raw_command('configure')
        if len(result.xpath('/nc:rpc-reply/nc:ok', namespaces=_NS_MAP)) > 0:
            logging.info('Command successful')
        else:
            logging.info('Command unsuccessful')

    except RPCError as err:
        logging.info('Error: %s' % err.message.strip())

    m.close_session()

if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s'
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=LOG_FORMAT)

    try:
        connect(sys.argv[1], '830', 'admin', 'admin')

    except IndexError:
        logging.error('Must supply a valid hostname or IP address')





import unittest
from ncclient.devices.csr import *


class TestCsrDevice(unittest.TestCase):
    
    def setUp(self):
        self.obj = CsrDeviceHandler({'name': 'csr'})

    def test_add_additional_ssh_connect_params(self):
        expected = dict()
        expected["unknown_host_cb"] = csr_unknown_host_cb
        actual = dict()
        self.obj.add_additional_ssh_connect_params(actual)
        self.assertDictEqual(expected, actual)

    def test_csr_unknown_host_cb(self):
        self.assertTrue(csr_unknown_host_cb('host', 'fingerprint'))
